from typing import TYPE_CHECKING, Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.job_metrics_executions_total import JobMetricsExecutionsTotal
    from ..models.job_metrics_tasks_total import JobMetricsTasksTotal
    from ..models.jobs_chart_value import JobsChartValue
    from ..models.jobs_success_failed_chart import JobsSuccessFailedChart


T = TypeVar("T", bound="JobMetrics")


@_attrs_define
class JobMetrics:
    """Metrics for job

    Attributes:
        billable_time (Union[Unset, list['JobsChartValue']]): Billable time
        cpu_usage (Union[Unset, list['JobsChartValue']]): CPU usage
        executions_chart (Union[Unset, list['JobsSuccessFailedChart']]): Executions chart
        executions_running (Union[Unset, list['JobsChartValue']]): Executions running
        executions_total (Union[Unset, JobMetricsExecutionsTotal]): Total executions
        ram_usage (Union[Unset, list['JobsChartValue']]): RAM usage
        tasks_chart (Union[Unset, list['JobsSuccessFailedChart']]): Tasks chart
        tasks_running (Union[Unset, list['JobsChartValue']]): Tasks running
        tasks_total (Union[Unset, JobMetricsTasksTotal]): Total tasks
    """

    billable_time: Union[Unset, list["JobsChartValue"]] = UNSET
    cpu_usage: Union[Unset, list["JobsChartValue"]] = UNSET
    executions_chart: Union[Unset, list["JobsSuccessFailedChart"]] = UNSET
    executions_running: Union[Unset, list["JobsChartValue"]] = UNSET
    executions_total: Union[Unset, "JobMetricsExecutionsTotal"] = UNSET
    ram_usage: Union[Unset, list["JobsChartValue"]] = UNSET
    tasks_chart: Union[Unset, list["JobsSuccessFailedChart"]] = UNSET
    tasks_running: Union[Unset, list["JobsChartValue"]] = UNSET
    tasks_total: Union[Unset, "JobMetricsTasksTotal"] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        billable_time: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.billable_time, Unset):
            billable_time = []
            for billable_time_item_data in self.billable_time:
                if type(billable_time_item_data) is dict:
                    billable_time_item = billable_time_item_data
                else:
                    billable_time_item = billable_time_item_data.to_dict()
                billable_time.append(billable_time_item)

        cpu_usage: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.cpu_usage, Unset):
            cpu_usage = []
            for cpu_usage_item_data in self.cpu_usage:
                if type(cpu_usage_item_data) is dict:
                    cpu_usage_item = cpu_usage_item_data
                else:
                    cpu_usage_item = cpu_usage_item_data.to_dict()
                cpu_usage.append(cpu_usage_item)

        executions_chart: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.executions_chart, Unset):
            executions_chart = []
            for executions_chart_item_data in self.executions_chart:
                if type(executions_chart_item_data) is dict:
                    executions_chart_item = executions_chart_item_data
                else:
                    executions_chart_item = executions_chart_item_data.to_dict()
                executions_chart.append(executions_chart_item)

        executions_running: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.executions_running, Unset):
            executions_running = []
            for executions_running_item_data in self.executions_running:
                if type(executions_running_item_data) is dict:
                    executions_running_item = executions_running_item_data
                else:
                    executions_running_item = executions_running_item_data.to_dict()
                executions_running.append(executions_running_item)

        executions_total: Union[Unset, dict[str, Any]] = UNSET
        if (
            self.executions_total
            and not isinstance(self.executions_total, Unset)
            and not isinstance(self.executions_total, dict)
        ):
            executions_total = self.executions_total.to_dict()
        elif self.executions_total and isinstance(self.executions_total, dict):
            executions_total = self.executions_total

        ram_usage: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.ram_usage, Unset):
            ram_usage = []
            for ram_usage_item_data in self.ram_usage:
                if type(ram_usage_item_data) is dict:
                    ram_usage_item = ram_usage_item_data
                else:
                    ram_usage_item = ram_usage_item_data.to_dict()
                ram_usage.append(ram_usage_item)

        tasks_chart: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.tasks_chart, Unset):
            tasks_chart = []
            for tasks_chart_item_data in self.tasks_chart:
                if type(tasks_chart_item_data) is dict:
                    tasks_chart_item = tasks_chart_item_data
                else:
                    tasks_chart_item = tasks_chart_item_data.to_dict()
                tasks_chart.append(tasks_chart_item)

        tasks_running: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.tasks_running, Unset):
            tasks_running = []
            for tasks_running_item_data in self.tasks_running:
                if type(tasks_running_item_data) is dict:
                    tasks_running_item = tasks_running_item_data
                else:
                    tasks_running_item = tasks_running_item_data.to_dict()
                tasks_running.append(tasks_running_item)

        tasks_total: Union[Unset, dict[str, Any]] = UNSET
        if self.tasks_total and not isinstance(self.tasks_total, Unset) and not isinstance(self.tasks_total, dict):
            tasks_total = self.tasks_total.to_dict()
        elif self.tasks_total and isinstance(self.tasks_total, dict):
            tasks_total = self.tasks_total

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if billable_time is not UNSET:
            field_dict["billableTime"] = billable_time
        if cpu_usage is not UNSET:
            field_dict["cpuUsage"] = cpu_usage
        if executions_chart is not UNSET:
            field_dict["executionsChart"] = executions_chart
        if executions_running is not UNSET:
            field_dict["executionsRunning"] = executions_running
        if executions_total is not UNSET:
            field_dict["executionsTotal"] = executions_total
        if ram_usage is not UNSET:
            field_dict["ramUsage"] = ram_usage
        if tasks_chart is not UNSET:
            field_dict["tasksChart"] = tasks_chart
        if tasks_running is not UNSET:
            field_dict["tasksRunning"] = tasks_running
        if tasks_total is not UNSET:
            field_dict["tasksTotal"] = tasks_total

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        from ..models.job_metrics_executions_total import JobMetricsExecutionsTotal
        from ..models.job_metrics_tasks_total import JobMetricsTasksTotal
        from ..models.jobs_chart_value import JobsChartValue
        from ..models.jobs_success_failed_chart import JobsSuccessFailedChart

        if not src_dict:
            return None
        d = src_dict.copy()
        billable_time = []
        _billable_time = d.pop("billableTime", UNSET)
        for billable_time_item_data in _billable_time or []:
            billable_time_item = JobsChartValue.from_dict(billable_time_item_data)

            billable_time.append(billable_time_item)

        cpu_usage = []
        _cpu_usage = d.pop("cpuUsage", UNSET)
        for cpu_usage_item_data in _cpu_usage or []:
            cpu_usage_item = JobsChartValue.from_dict(cpu_usage_item_data)

            cpu_usage.append(cpu_usage_item)

        executions_chart = []
        _executions_chart = d.pop("executionsChart", UNSET)
        for executions_chart_item_data in _executions_chart or []:
            executions_chart_item = JobsSuccessFailedChart.from_dict(executions_chart_item_data)

            executions_chart.append(executions_chart_item)

        executions_running = []
        _executions_running = d.pop("executionsRunning", UNSET)
        for executions_running_item_data in _executions_running or []:
            executions_running_item = JobsChartValue.from_dict(executions_running_item_data)

            executions_running.append(executions_running_item)

        _executions_total = d.pop("executionsTotal", UNSET)
        executions_total: Union[Unset, JobMetricsExecutionsTotal]
        if isinstance(_executions_total, Unset):
            executions_total = UNSET
        else:
            executions_total = JobMetricsExecutionsTotal.from_dict(_executions_total)

        ram_usage = []
        _ram_usage = d.pop("ramUsage", UNSET)
        for ram_usage_item_data in _ram_usage or []:
            ram_usage_item = JobsChartValue.from_dict(ram_usage_item_data)

            ram_usage.append(ram_usage_item)

        tasks_chart = []
        _tasks_chart = d.pop("tasksChart", UNSET)
        for tasks_chart_item_data in _tasks_chart or []:
            tasks_chart_item = JobsSuccessFailedChart.from_dict(tasks_chart_item_data)

            tasks_chart.append(tasks_chart_item)

        tasks_running = []
        _tasks_running = d.pop("tasksRunning", UNSET)
        for tasks_running_item_data in _tasks_running or []:
            tasks_running_item = JobsChartValue.from_dict(tasks_running_item_data)

            tasks_running.append(tasks_running_item)

        _tasks_total = d.pop("tasksTotal", UNSET)
        tasks_total: Union[Unset, JobMetricsTasksTotal]
        if isinstance(_tasks_total, Unset):
            tasks_total = UNSET
        else:
            tasks_total = JobMetricsTasksTotal.from_dict(_tasks_total)

        job_metrics = cls(
            billable_time=billable_time,
            cpu_usage=cpu_usage,
            executions_chart=executions_chart,
            executions_running=executions_running,
            executions_total=executions_total,
            ram_usage=ram_usage,
            tasks_chart=tasks_chart,
            tasks_running=tasks_running,
            tasks_total=tasks_total,
        )

        job_metrics.additional_properties = d
        return job_metrics

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import io
import json
import threading
from pathlib import Path
from typing import Any, Callable, Dict, List, Union

import httpx

from ...common.settings import settings
from ..client.models import Directory, FileRequest, SuccessResponse
from ..types import (
    CopyResponse,
    SandboxConfiguration,
    SandboxFilesystemFile,
    WatchEvent,
)
from .action import SyncSandboxAction

# Multipart upload constants
MULTIPART_THRESHOLD = 5 * 1024 * 1024  # 5MB
CHUNK_SIZE = 5 * 1024 * 1024  # 5MB per part
MAX_PARALLEL_UPLOADS = 3  # Number of parallel part uploads


class SyncSandboxFileSystem(SyncSandboxAction):
    def __init__(self, sandbox_config: SandboxConfiguration, process=None):
        super().__init__(sandbox_config)
        self.process = process

    def mkdir(self, path: str, permissions: str = "0755") -> SuccessResponse:
        path = self.format_path(path)
        body = FileRequest(is_directory=True, permissions=permissions)
        with self.get_client() as client_instance:
            response = client_instance.put(f"/filesystem/{path}", json=body.to_dict())
            self.handle_response_error(response)
            return SuccessResponse.from_dict(response.json())

    def write(self, path: str, content: str) -> SuccessResponse:
        path = self.format_path(path)
        content_size = len(content.encode("utf-8"))
        if content_size > MULTIPART_THRESHOLD:
            content_bytes = content.encode("utf-8")
            return self._upload_with_multipart(path, content_bytes, "0644")
        body = FileRequest(content=content)
        with self.get_client() as client_instance:
            response = client_instance.put(f"/filesystem/{path}", json=body.to_dict())
            self.handle_response_error(response)
            return SuccessResponse.from_dict(response.json())

    def write_binary(self, path: str, content: Union[bytes, bytearray, str]) -> SuccessResponse:
        path = self.format_path(path)
        if isinstance(content, str):
            local_path = Path(content)
            content = local_path.read_bytes()
        elif isinstance(content, bytearray):
            content = bytes(content)
        if len(content) > MULTIPART_THRESHOLD:
            return self._upload_with_multipart(path, content, "0644")
        binary_file = io.BytesIO(content)
        files = {
            "file": ("binary-file.bin", binary_file, "application/octet-stream"),
        }
        data = {"permissions": "0644", "path": path}
        url = f"{self.url}/filesystem/{path}"
        headers = {**settings.headers, **self.sandbox_config.headers}
        with self.get_client() as client_instance:
            response = client_instance.put(url, files=files, data=data, headers=headers)
            if not response.is_success:
                raise Exception(f"Failed to write binary: {response.status_code} {response.text}")
            return SuccessResponse.from_dict(response.json())

    def write_tree(
        self,
        files: List[Union[SandboxFilesystemFile, Dict[str, Any]]],
        destination_path: str | None = None,
    ) -> Directory:
        files_dict = {}
        for file in files:
            if isinstance(file, dict):
                file = SandboxFilesystemFile.from_dict(file)
            files_dict[file.path] = file.content
        path = destination_path or ""
        with self.get_client() as client_instance:
            response = client_instance.put(
                f"/filesystem/tree/{path}",
                json={"files": files_dict},
                headers={"Content-Type": "application/json"},
            )
            self.handle_response_error(response)
            return Directory.from_dict(response.json())

    def read(self, path: str) -> str:
        path = self.format_path(path)
        with self.get_client() as client_instance:
            response = client_instance.get(f"/filesystem/{path}")
            self.handle_response_error(response)
            data = response.json()
            if "content" in data:
                return data["content"]
            raise Exception("Unsupported file type")

    def read_binary(self, path: str) -> bytes:
        path = self.format_path(path)
        url = f"{self.url}/filesystem/{path}"
        headers = {
            **settings.headers,
            **self.sandbox_config.headers,
            "Accept": "application/octet-stream",
        }
        with self.get_client() as client_instance:
            response = client_instance.get(url, headers=headers)
            self.handle_response_error(response)
            return response.content

    def download(self, src: str, destination_path: str, mode: int = 0o644) -> None:
        content = self.read_binary(src)
        local_path = Path(destination_path)
        local_path.write_bytes(content)
        local_path.chmod(mode)

    def rm(self, path: str, recursive: bool = False) -> SuccessResponse:
        path = self.format_path(path)
        with self.get_client() as client_instance:
            params = {"recursive": "true"} if recursive else {}
            response = client_instance.delete(f"/filesystem/{path}", params=params)
            self.handle_response_error(response)
            return SuccessResponse.from_dict(response.json())

    def ls(self, path: str) -> Directory:
        path = self.format_path(path)
        with self.get_client() as client_instance:
            response = client_instance.get(f"/filesystem/{path}")
            self.handle_response_error(response)
            data = response.json()
            if not ("files" in data or "subdirectories" in data):
                raise Exception('{"error": "Directory not found"}')
            return Directory.from_dict(data)

    def cp(self, source: str, destination: str, max_wait: int = 180000) -> CopyResponse:
        if not self.process:
            raise Exception("Process instance not available. Cannot execute cp command.")
        process = self.process.exec({"command": f"cp -r {source} {destination}"})
        process = self.process.wait(process.pid, max_wait=max_wait, interval=100)
        if process.status == "failed":
            logs = process.logs if hasattr(process, "logs") else "Unknown error"
            raise Exception(f"Could not copy {source} to {destination} cause: {logs}")
        return CopyResponse(
            message="Files copied",
            source=source,
            destination=destination,
        )

    def watch(
        self,
        path: str,
        callback: Callable[[WatchEvent], None],
        options: Dict[str, Any] | None = None,
    ) -> Dict[str, Callable]:
        path = self.format_path(path)
        closed = threading.Event()
        if options is None:
            options = {}

        def run():
            params = {}
            if options.get("ignore"):
                params["ignore"] = ",".join(options["ignore"])
            url = f"{self.url}/watch/filesystem/{path}"
            headers = {**settings.headers, **self.sandbox_config.headers}
            with httpx.Client() as client_instance:
                with client_instance.stream("GET", url, params=params, headers=headers) as response:
                    if not response.is_success:
                        raise Exception(f"Failed to start watching: {response.status_code}")
                    buffer = ""
                    for chunk in response.iter_text():
                        if closed.is_set():
                            break
                        buffer += chunk
                        lines = buffer.split("\n")
                        buffer = lines.pop()  # Keep incomplete line in buffer
                        for line in lines:
                            line = line.strip()
                            if not line:
                                continue
                            if line.startswith("[keepalive]"):
                                continue
                            try:
                                file_event_data = json.loads(line)
                                file_event = WatchEvent(
                                    op=file_event_data.get("op", ""),
                                    path=file_event_data.get("path", ""),
                                    name=file_event_data.get("name", ""),
                                    content=file_event_data.get("content"),
                                )
                                if options.get("with_content") and file_event.op in ["CREATE", "WRITE"]:
                                    try:
                                        file_path = file_event.path
                                        if file_path.endswith("/"):
                                            file_path = file_path + file_event.name
                                        else:
                                            file_path = file_path + "/" + file_event.name
                                        content = self.read(file_path)
                                        file_event.content = content
                                    except Exception:
                                        file_event.content = None
                                callback(file_event)
                            except json.JSONDecodeError:
                                continue
                            except Exception as e:
                                if options.get("on_error"):
                                    options["on_error"](e)

        thread = threading.Thread(target=run, daemon=True)
        thread.start()

        def close():
            closed.set()

        return {"close": close}

    def format_path(self, path: str) -> str:
        return path

    def _initiate_multipart_upload(self, path: str, permissions: str = "0644") -> Dict[str, Any]:
        path = self.format_path(path)
        url = f"{self.url}/filesystem-multipart/initiate/{path}"
        headers = {**settings.headers, **self.sandbox_config.headers}
        body = {"permissions": permissions}
        with self.get_client() as client_instance:
            response = client_instance.post(url, json=body, headers=headers)
            self.handle_response_error(response)
            return response.json()

    def _upload_part(self, upload_id: str, part_number: int, data: bytes) -> Dict[str, Any]:
        url = f"{self.url}/filesystem-multipart/{upload_id}/part"
        headers = {**settings.headers, **self.sandbox_config.headers}
        params = {"partNumber": part_number}
        files = {"file": ("part", io.BytesIO(data), "application/octet-stream")}
        with self.get_client() as client_instance:
            response = client_instance.put(url, files=files, params=params, headers=headers)
            self.handle_response_error(response)
            return response.json()

    def _complete_multipart_upload(self, upload_id: str, parts: List[Dict[str, Any]]) -> SuccessResponse:
        url = f"{self.url}/filesystem-multipart/{upload_id}/complete"
        headers = {**settings.headers, **self.sandbox_config.headers}
        body = {"parts": parts}
        with self.get_client() as client_instance:
            response = client_instance.post(url, json=body, headers=headers)
            self.handle_response_error(response)
            return SuccessResponse.from_dict(response.json())

    def _abort_multipart_upload(self, upload_id: str) -> None:
        url = f"{self.url}/filesystem-multipart/{upload_id}/abort"
        headers = {**settings.headers, **self.sandbox_config.headers}
        with self.get_client() as client_instance:
            response = client_instance.delete(url, headers=headers)
            if not response.is_success:
                print(f"Warning: Failed to abort multipart upload: {response.status_code}")

    def _upload_with_multipart(self, path: str, data: bytes, permissions: str = "0644") -> SuccessResponse:
        init_response = self._initiate_multipart_upload(path, permissions)
        upload_id = init_response.get("uploadId")
        if not upload_id:
            raise Exception("Failed to get upload ID from initiate response")
        try:
            size = len(data)
            num_parts = (size + CHUNK_SIZE - 1) // CHUNK_SIZE
            parts: List[Dict[str, Any]] = []
            # Upload parts sequentially but in groups using threads for parallelism
            for i in range(0, num_parts, MAX_PARALLEL_UPLOADS):
                threads = []
                results: Dict[int, Dict[str, Any]] = {}
                def make_upload(part_number: int, chunk: bytes):
                    results[part_number] = self._upload_part(upload_id, part_number, chunk)
                for j in range(MAX_PARALLEL_UPLOADS):
                    if i + j >= num_parts:
                        break
                    part_number = i + j + 1
                    start = (part_number - 1) * CHUNK_SIZE
                    end = min(start + CHUNK_SIZE, size)
                    chunk = data[start:end]
                    t = threading.Thread(target=make_upload, args=(part_number, chunk))
                    threads.append(t)
                    t.start()
                for t in threads:
                    t.join()
                for part_number, r in results.items():
                    parts.append({"partNumber": part_number, "etag": r.get("etag")})
            parts.sort(key=lambda p: p.get("partNumber", 0))
            return self._complete_multipart_upload(upload_id, parts)
        except Exception as error:
            try:
                self._abort_multipart_upload(upload_id)
            except Exception as abort_error:
                print(f"Failed to abort multipart upload: {abort_error}")
            raise error





from typing import TYPE_CHECKING, Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.filesystem_multipart_upload_parts import FilesystemMultipartUploadParts


T = TypeVar("T", bound="FilesystemMultipartUpload")


@_attrs_define
class FilesystemMultipartUpload:
    """
    Attributes:
        initiated_at (Union[Unset, str]):
        parts (Union[Unset, FilesystemMultipartUploadParts]):
        path (Union[Unset, str]):  Example: /tmp/largefile.dat.
        permissions (Union[Unset, int]):  Example: 420.
        upload_id (Union[Unset, str]):  Example: 550e8400-e29b-41d4-a716-446655440000.
    """

    initiated_at: Union[Unset, str] = UNSET
    parts: Union[Unset, "FilesystemMultipartUploadParts"] = UNSET
    path: Union[Unset, str] = UNSET
    permissions: Union[Unset, int] = UNSET
    upload_id: Union[Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        initiated_at = self.initiated_at

        parts: Union[Unset, dict[str, Any]] = UNSET
        if self.parts and not isinstance(self.parts, Unset) and not isinstance(self.parts, dict):
            parts = self.parts.to_dict()
        elif self.parts and isinstance(self.parts, dict):
            parts = self.parts

        path = self.path

        permissions = self.permissions

        upload_id = self.upload_id

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if initiated_at is not UNSET:
            field_dict["initiatedAt"] = initiated_at
        if parts is not UNSET:
            field_dict["parts"] = parts
        if path is not UNSET:
            field_dict["path"] = path
        if permissions is not UNSET:
            field_dict["permissions"] = permissions
        if upload_id is not UNSET:
            field_dict["uploadId"] = upload_id

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        from ..models.filesystem_multipart_upload_parts import FilesystemMultipartUploadParts

        if not src_dict:
            return None
        d = src_dict.copy()
        initiated_at = d.pop("initiatedAt", UNSET)

        _parts = d.pop("parts", UNSET)
        parts: Union[Unset, FilesystemMultipartUploadParts]
        if isinstance(_parts, Unset):
            parts = UNSET
        else:
            parts = FilesystemMultipartUploadParts.from_dict(_parts)

        path = d.pop("path", UNSET)

        permissions = d.pop("permissions", UNSET)

        upload_id = d.pop("uploadId", UNSET)

        filesystem_multipart_upload = cls(
            initiated_at=initiated_at,
            parts=parts,
            path=path,
            permissions=permissions,
            upload_id=upload_id,
        )

        filesystem_multipart_upload.additional_properties = d
        return filesystem_multipart_upload

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from http import HTTPStatus
from typing import Any, Union, cast

import httpx

from ... import errors
from ...client import Client
from ...models.job_execution import JobExecution
from ...types import UNSET, Response, Unset


def _get_kwargs(
    job_id: str,
    *,
    limit: Union[Unset, int] = 20,
    offset: Union[Unset, int] = 0,
) -> dict[str, Any]:
    params: dict[str, Any] = {}

    params["limit"] = limit

    params["offset"] = offset

    params = {k: v for k, v in params.items() if v is not UNSET and v is not None}

    _kwargs: dict[str, Any] = {
        "method": "get",
        "url": f"/jobs/{job_id}/executions",
        "params": params,
    }

    return _kwargs


def _parse_response(*, client: Client, response: httpx.Response) -> Union[Any, list["JobExecution"]] | None:
    if response.status_code == 200:
        response_200 = []
        _response_200 = response.json()
        for response_200_item_data in _response_200:
            response_200_item = JobExecution.from_dict(response_200_item_data)

            response_200.append(response_200_item)

        return response_200
    if response.status_code == 400:
        response_400 = cast(Any, None)
        return response_400
    if response.status_code == 500:
        response_500 = cast(Any, None)
        return response_500
    if client.raise_on_unexpected_status:
        raise errors.UnexpectedStatus(response.status_code, response.content)
    else:
        return None


def _build_response(*, client: Client, response: httpx.Response) -> Response[Union[Any, list["JobExecution"]]]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(client=client, response=response),
    )


def sync_detailed(
    job_id: str,
    *,
    client: Union[Client],
    limit: Union[Unset, int] = 20,
    offset: Union[Unset, int] = 0,
) -> Response[Union[Any, list["JobExecution"]]]:
    """List job executions

     Returns a list of all executions for a job by name.

    Args:
        job_id (str):
        limit (Union[Unset, int]):  Default: 20.
        offset (Union[Unset, int]):  Default: 0.

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[Any, list['JobExecution']]]
    """

    kwargs = _get_kwargs(
        job_id=job_id,
        limit=limit,
        offset=offset,
    )

    response = client.get_httpx_client().request(
        **kwargs,
    )

    return _build_response(client=client, response=response)


def sync(
    job_id: str,
    *,
    client: Union[Client],
    limit: Union[Unset, int] = 20,
    offset: Union[Unset, int] = 0,
) -> Union[Any, list["JobExecution"]] | None:
    """List job executions

     Returns a list of all executions for a job by name.

    Args:
        job_id (str):
        limit (Union[Unset, int]):  Default: 20.
        offset (Union[Unset, int]):  Default: 0.

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[Any, list['JobExecution']]
    """

    return sync_detailed(
        job_id=job_id,
        client=client,
        limit=limit,
        offset=offset,
    ).parsed


async def asyncio_detailed(
    job_id: str,
    *,
    client: Union[Client],
    limit: Union[Unset, int] = 20,
    offset: Union[Unset, int] = 0,
) -> Response[Union[Any, list["JobExecution"]]]:
    """List job executions

     Returns a list of all executions for a job by name.

    Args:
        job_id (str):
        limit (Union[Unset, int]):  Default: 20.
        offset (Union[Unset, int]):  Default: 0.

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[Any, list['JobExecution']]]
    """

    kwargs = _get_kwargs(
        job_id=job_id,
        limit=limit,
        offset=offset,
    )

    response = await client.get_async_httpx_client().request(**kwargs)

    return _build_response(client=client, response=response)


async def asyncio(
    job_id: str,
    *,
    client: Union[Client],
    limit: Union[Unset, int] = 20,
    offset: Union[Unset, int] = 0,
) -> Union[Any, list["JobExecution"]] | None:
    """List job executions

     Returns a list of all executions for a job by name.

    Args:
        job_id (str):
        limit (Union[Unset, int]):  Default: 20.
        offset (Union[Unset, int]):  Default: 0.

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[Any, list['JobExecution']]
    """

    return (
        await asyncio_detailed(
            job_id=job_id,
            client=client,
            limit=limit,
            offset=offset,
        )
    ).parsed



import asyncio
import os
import sys

from blaxel.core.sandbox import SandboxInstance

# Determine base image based on environment
base_image = "blaxel/vite:latest" if os.getenv("BL_ENV") == "dev" else "blaxel/prod-base:latest"
print(f"Using base image: {base_image} (BL_ENV={os.getenv('BL_ENV', 'not set')})\n")


async def wait_for_termination(sandbox_name: str, max_minutes: int = 10) -> bool:
    """Wait for sandbox to be terminated, checking every minute."""
    print(f"Checking sandbox status every minute for up to {max_minutes} minutes...")

    for minute in range(1, max_minutes + 1):
        # Wait 1 minute
        await asyncio.sleep(60)

        # Check sandbox status
        sandbox_status = await SandboxInstance.get(sandbox_name)
        print(f"  Minute {minute}: Status = {sandbox_status.status}")

        if sandbox_status.status in ["TERMINATED", "DELETED"]:
            print(f"✅ Sandbox terminated after {minute} minute(s)")
            return True

    return False


async def main():
    """Test expiration policies with ttl-idle and createIfNotExists."""
    try:
        sandbox_name = "sandbox-ttl-idle"

        print("=== Test 1: Create sandbox with ttl-idle expiration policy ===")

        # Create sandbox with ttl-idle expiration policy of 20 seconds
        sandbox = await SandboxInstance.create_if_not_exists({
            "name": sandbox_name,
            "image": base_image,
            "lifecycle": {
                "expirationPolicies": [
                    {
                        "type": "ttl-idle",
                        "value": "20s",
                        "action": "delete",
                    },
                ],
            },
        })

        print(f"✅ Created sandbox with ttl-idle policy: {sandbox.metadata.name}")

        # IMPORTANT: Make a sandbox API call to activate idle monitoring for ttl-idle policy
        print("Making sandbox API call to activate idle monitoring...")
        exec_result = await sandbox.process.exec({"command": "echo 'activate idle timer'"})
        print(f"Process exec successful (pid: {exec_result.pid}) - Idle timer activated")

        # Test 1.5: Try to create_if_not_exists with same name while running
        print("\n=== Test 1.5: Attempt to create_if_not_exists with same name while running ===")
        print("Attempting to create_if_not_exists with duplicate name (should return existing sandbox)...")

        duplicate_sandbox = await SandboxInstance.create_if_not_exists({
            "name": sandbox_name,
            "image": base_image,
            "lifecycle": {
                "expirationPolicies": [
                    {
                        "type": "ttl-idle",
                        "value": "20s",
                        "action": "delete",
                    },
                ],
            },
        })

        # With create_if_not_exists, we should get the existing active sandbox
        if duplicate_sandbox.metadata.name == sandbox_name and duplicate_sandbox.status in ["RUNNING", "DEPLOYED"]:
            print("✅ Test 1.5 PASSED: create_if_not_exists correctly returned the existing active sandbox")
            print(f"  Returned sandbox: {duplicate_sandbox.metadata.name}, Status: {duplicate_sandbox.status}")
        else:
            print("❌ Test 1.5 FAILED: create_if_not_exists did not return the expected active sandbox")
            print(f"  Got: {duplicate_sandbox.metadata.name}, Status: {duplicate_sandbox.status}")
            sys.exit(1)

        print("\nContinuing with original sandbox termination test...")

        # Wait for termination (checking every minute for up to 10 minutes)
        terminated1 = await wait_for_termination(sandbox_name, 10)

        if terminated1:
            print("✅ Test 1 PASSED: Sandbox was terminated due to idle timeout.")
        else:
            print("❌ Test 1 FAILED: Sandbox was not terminated within 10 minutes.")

            # Clean up if not terminated
            final_status = await SandboxInstance.get(sandbox_name)
            if final_status.status in ["RUNNING", "DEPLOYED"]:
                await SandboxInstance.delete(sandbox_name)
                print("Cleaned up sandbox manually.")
            sys.exit(1)

        print("\n=== Test 2: Use create_if_not_exists with same name after termination ===")
        print("Using create_if_not_exists with the same name - should automatically recreate since previous is TERMINATED...")

        # Create another sandbox with the same name
        sandbox = await SandboxInstance.create_if_not_exists({
            "name": sandbox_name,
            "image": base_image,
            "lifecycle": {
                "expirationPolicies": [
                    {
                        "type": "ttl-idle",
                        "value": "20s",
                        "action": "delete",
                    },
                ],
            },
        })

        print(f"✅ Successfully created new sandbox with same name: {sandbox.metadata.name}")

        # Activate idle monitoring again
        print("Making sandbox API call to activate idle monitoring...")
        exec_result = await sandbox.process.exec({"command": "echo 'activate idle timer for second sandbox'"})
        print(f"Process exec successful (pid: {exec_result.pid}) - Idle timer activated")

        # Wait for second sandbox termination
        terminated2 = await wait_for_termination(sandbox_name, 10)

        if terminated2:
            print("✅ Test 2 PASSED: Second sandbox was also terminated due to idle timeout.")
            print("\n✅✅✅ ALL TESTS PASSED ✅✅✅")
        else:
            print("❌ Test 2 FAILED: Second sandbox was not terminated within 10 minutes.")

            # Clean up if not terminated
            final_status = await SandboxInstance.get(sandbox_name)
            if final_status.status in ["RUNNING", "DEPLOYED"]:
                await SandboxInstance.delete(sandbox_name)
                print("Cleaned up sandbox manually.")
            sys.exit(1)

    except Exception as e:
        print(f"❌ There was an error => {e}")
        import traceback
        traceback.print_exc()
        sys.exit(1)


if __name__ == "__main__":
    try:
        asyncio.run(main())
        sys.exit(0)
    except Exception as err:
        print(f"❌ There was an error => {err}")
        sys.exit(1)




from typing import TYPE_CHECKING, Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.expiration_policy import ExpirationPolicy


T = TypeVar("T", bound="SandboxLifecycle")


@_attrs_define
class SandboxLifecycle:
    """Lifecycle configuration for sandbox management

    Attributes:
        expiration_policies (Union[Unset, list['ExpirationPolicy']]): List of expiration policies
    """

    expiration_policies: Union[Unset, list["ExpirationPolicy"]] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        expiration_policies: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.expiration_policies, Unset):
            expiration_policies = []
            for expiration_policies_item_data in self.expiration_policies:
                if type(expiration_policies_item_data) is dict:
                    expiration_policies_item = expiration_policies_item_data
                else:
                    expiration_policies_item = expiration_policies_item_data.to_dict()
                expiration_policies.append(expiration_policies_item)

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if expiration_policies is not UNSET:
            field_dict["expirationPolicies"] = expiration_policies

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        from ..models.expiration_policy import ExpirationPolicy

        if not src_dict:
            return None
        d = src_dict.copy()
        expiration_policies = []
        _expiration_policies = d.pop("expirationPolicies", UNSET)
        for expiration_policies_item_data in _expiration_policies or []:
            expiration_policies_item = ExpirationPolicy.from_dict(expiration_policies_item_data)

            expiration_policies.append(expiration_policies_item)

        sandbox_lifecycle = cls(
            expiration_policies=expiration_policies,
        )

        sandbox_lifecycle.additional_properties = d
        return sandbox_lifecycle

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from __future__ import annotations

import asyncio
import json
import logging
import uuid
import warnings
from operator import itemgetter
from typing import (
    Any,
    AsyncIterator,
    Callable,
    Dict,
    Iterator,
    List,
    Mapping,
    Sequence,
    Tuple,
    Type,
    Union,
    cast,
)

import httpx
import requests
from langchain_core.callbacks.manager import AsyncCallbackManagerForLLMRun, CallbackManagerForLLMRun
from langchain_core.language_models import LanguageModelInput
from langchain_core.language_models.chat_models import BaseChatModel, LangSmithParams
from langchain_core.messages import (
    AIMessage,
    AIMessageChunk,
    BaseMessage,
    FunctionMessage,
    HumanMessage,
    SystemMessage,
    ToolMessage,
)
from langchain_core.messages.ai import UsageMetadata
from langchain_core.messages.tool import invalid_tool_call, tool_call, tool_call_chunk
from langchain_core.output_parsers.openai_tools import (
    JsonOutputKeyToolsParser,
    PydanticToolsParser,
    parse_tool_calls,
)
from langchain_core.outputs import ChatGeneration, ChatGenerationChunk, ChatResult
from langchain_core.runnables import Runnable, RunnablePassthrough
from langchain_core.tools import BaseTool
from langchain_core.utils.function_calling import convert_to_openai_tool
from PIL import Image
from pydantic import BaseModel, ConfigDict, Field, SecretStr, model_validator
from tenacity import (
    before_sleep_log,
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    wait_exponential,
)
from typing_extensions import Self, is_typeddict

WARNED_STRUCTURED_OUTPUT_JSON_MODE = False

logger = logging.getLogger(__name__)

# Type definitions
_FunctionDeclarationType = Union[
    Dict[str, Any],
    Callable[..., Any],
]

_ToolChoiceType = Union[str, bool, Dict[str, Any]]
_ToolConfigDict = Dict[str, Any]
_ToolDict = Dict[str, Any]
SafetySettingDict = Dict[str, str]
OutputParserLike = Union[PydanticToolsParser, JsonOutputKeyToolsParser]

# Data classes
class Part(BaseModel):
    text: str | None = None
    inline_data: Dict[str, Any] | None = None
    file_data: Dict[str, Any] | None = None
    function_call: Dict[str, Any] | None = None
    function_response: Dict[str, Any] | None = None

class Content(BaseModel):
    role: str | None = None
    parts: List[Part]

class Blob(BaseModel):
    data: str
    mime_type: str

class FileData(BaseModel):
    file_uri: str
    mime_type: str

class VideoMetadata(BaseModel):
    duration: str | None = None
    start_offset: str | None = None
    end_offset: str | None = None

class FunctionCall(BaseModel):
    name: str
    args: Dict[str, Any]

class FunctionResponse(BaseModel):
    name: str
    response: Dict[str, Any]

class SafetySetting(BaseModel):
    category: str
    threshold: str

class ToolConfig(BaseModel):
    function_calling_config: Dict[str, Any]

class GenerationConfig(BaseModel):
    candidate_count: int | None = None
    temperature: float | None = None
    stop_sequences: List[str] | None = None
    max_output_tokens: int | None = None
    top_k: int | None = None
    top_p: float | None = None
    response_modalities: List[str] | None = None

class GoogleTool(BaseModel):
    name: str
    description: str
    parameters: Dict[str, Any]

class ImageBytesLoader:
    def load_part(self, image_url: str) -> Part:
        """Load an image from a URL and convert it to a Part."""
        import io

        response = requests.get(image_url)
        response.raise_for_status()

        # Convert to JPEG format
        img = Image.open(io.BytesIO(response.content))
        img_byte_arr = io.BytesIO()
        img.save(img_byte_arr, format='JPEG')
        img_byte_arr = img_byte_arr.getvalue()

        return Part(
            inline_data={
                "mime_type": "image/jpeg",
                "data": img_byte_arr.decode('utf-8')
            }
        )

class ChatGoogleGenerativeAIError(Exception):
    """
    Custom exception class for errors associated with the `Google GenAI` API.

    This exception is raised when there are specific issues related to the
    Google genai API usage in the ChatGoogleGenerativeAI class, such as unsupported
    message types or roles.
    """
    pass


def _create_retry_decorator() -> Callable[[Any], Any]:
    """
    Creates and returns a preconfigured tenacity retry decorator.

    The retry decorator is configured to handle specific HTTP exceptions
    such as rate limits and service unavailability. It uses an exponential
    backoff strategy for retries.

    Returns:
        Callable[[Any], Any]: A retry decorator configured for handling specific
        HTTP exceptions.
    """
    multiplier = 2
    min_seconds = 1
    max_seconds = 60
    max_retries = 2

    return retry(
        reraise=True,
        stop=stop_after_attempt(max_retries),
        wait=wait_exponential(multiplier=multiplier, min=min_seconds, max=max_seconds),
        retry=(
            retry_if_exception_type(httpx.HTTPStatusError)
            | retry_if_exception_type(httpx.ConnectError)
            | retry_if_exception_type(httpx.ReadTimeout)
        ),
        before_sleep=before_sleep_log(logger, logging.WARNING),
    )


def _chat_with_retry(generation_method: Callable, **kwargs: Any) -> Any:
    """
    Executes a chat generation method with retry logic using tenacity.

    This function is a wrapper that applies a retry mechanism to a provided
    chat generation function. It is useful for handling intermittent issues
    like network errors or temporary service unavailability.

    Args:
        generation_method (Callable): The chat generation method to be executed.
        **kwargs (Any): Additional keyword arguments to pass to the generation method.

    Returns:
        Any: The result from the chat generation method.
    """
    retry_decorator = _create_retry_decorator()

    @retry_decorator
    def _chat_with_retry(**kwargs: Any) -> Any:
        try:
            # Extract request parameters and other kwargs
            request = kwargs.pop('request', {})
            kwargs.pop('metadata', None)

            # Unpack request parameters into kwargs
            kwargs.update(request)

            return generation_method(**kwargs)
        except httpx.HTTPStatusError as e:
            if e.response.status_code == 429:  # Rate limit
                error_msg = (
                    "Rate limit exceeded. Please try again later or use a different API key."
                )
                raise ValueError(error_msg)
            elif e.response.status_code == 403:  # Forbidden
                error_msg = (
                    "Access forbidden. Please check your API key and permissions."
                )
                raise ValueError(error_msg)
            else:
                raise ChatGoogleGenerativeAIError(
                    f"HTTP error occurred: {e.response.text}"
                ) from e
        except Exception as e:
            raise e

    return _chat_with_retry(**kwargs)


async def _achat_with_retry(generation_method: Callable, **kwargs: Any) -> Any:
    """
    Executes a chat generation method with retry logic using tenacity.

    This function is a wrapper that applies a retry mechanism to a provided
    chat generation function. It is useful for handling intermittent issues
    like network errors or temporary service unavailability.

    Args:
        generation_method (Callable): The chat generation method to be executed.
        **kwargs (Any): Additional keyword arguments to pass to the generation method.

    Returns:
        Any: The result from the chat generation method.
    """
    retry_decorator = _create_retry_decorator()

    @retry_decorator
    async def _achat_with_retry(**kwargs: Any) -> Any:
        try:
            # Extract request parameters and other kwargs
            request = kwargs.pop('request', {})
            kwargs.pop('metadata', None)

            # Unpack request parameters into kwargs
            kwargs.update(request)

            return await generation_method(**kwargs)
        except httpx.HTTPStatusError as e:
            if e.response.status_code == 429:  # Rate limit
                error_msg = (
                    "Rate limit exceeded. Please try again later or use a different API key."
                )
                raise ValueError(error_msg)
            elif e.response.status_code == 403:  # Forbidden
                error_msg = (
                    "Access forbidden. Please check your API key and permissions."
                )
                raise ValueError(error_msg)
            else:
                raise ChatGoogleGenerativeAIError(
                    f"HTTP error occurred: {e.response.text}"
                ) from e
        except Exception as e:
            raise e

    return await _achat_with_retry(**kwargs)


def _is_openai_parts_format(part: dict) -> bool:
    return "type" in part


def _convert_to_parts(
    raw_content: Union[str, Sequence[Union[str, dict]]],
) -> List[Part]:
    """Converts a list of LangChain messages into a google parts."""
    parts = []
    content = [raw_content] if isinstance(raw_content, str) else raw_content
    image_loader = ImageBytesLoader()
    for part in content:
        if isinstance(part, str):
            parts.append(Part(text=part))
        elif isinstance(part, Mapping):
            # OpenAI Format
            if _is_openai_parts_format(part):
                if part["type"] == "text":
                    parts.append(Part(text=part["text"]))
                elif part["type"] == "image_url":
                    img_url = part["image_url"]
                    if isinstance(img_url, dict):
                        if "url" not in img_url:
                            raise ValueError(
                                f"Unrecognized message image format: {img_url}"
                            )
                        img_url = img_url["url"]
                    parts.append(image_loader.load_part(img_url))
                # Handle media type like LangChain.js
                # https://github.com/langchain-ai/langchainjs/blob/e536593e2585f1dd7b0afc187de4d07cb40689ba/libs/langchain-google-common/src/utils/gemini.ts#L93-L106
                elif part["type"] == "media":
                    if "mime_type" not in part:
                        raise ValueError(f"Missing mime_type in media part: {part}")
                    mime_type = part["mime_type"]
                    media_part = Part()

                    if "data" in part:
                        media_part.inline_data = Blob(
                            data=part["data"], mime_type=mime_type
                        )
                    elif "file_uri" in part:
                        media_part.file_data = FileData(
                            file_uri=part["file_uri"], mime_type=mime_type
                        )
                    else:
                        raise ValueError(
                            f"Media part must have either data or file_uri: {part}"
                        )
                    parts.append(media_part)
                else:
                    raise ValueError(
                        f"Unrecognized message part type: {part['type']}. Only text, "
                        f"image_url, and media types are supported."
                    )
            else:
                # Yolo
                logger.warning(
                    "Unrecognized message part format. Assuming it's a text part."
                )
                parts.append(Part(text=str(part)))
        else:
            # TODO: Maybe some of Google's native stuff
            # would hit this branch.
            raise ChatGoogleGenerativeAIError(
                "Gemini only supports text and inline_data parts."
            )
    return parts


def _convert_tool_message_to_part(
    message: ToolMessage | FunctionMessage, name: str | None = None
) -> Part:
    """Converts a tool or function message to a google part."""
    # Legacy agent stores tool name in message.additional_kwargs instead of message.name
    name = message.name or name or message.additional_kwargs.get("name")
    response: Any
    if not isinstance(message.content, str):
        response = message.content
    else:
        try:
            response = json.loads(message.content)
        except json.JSONDecodeError:
            response = message.content  # leave as str representation
    if not isinstance(response, dict):
        response = {"output": response}
    part = Part(
        function_response=FunctionResponse(
            name=name,
            response=response,
        ).model_dump()
    )
    return part


def _get_ai_message_tool_messages_parts(
    tool_messages: Sequence[ToolMessage], ai_message: AIMessage
) -> list[Part]:
    """
    Finds relevant tool messages for the AI message and converts them to a single
    list of Parts.
    """
    # We are interested only in the tool messages that are part of the AI message
    tool_calls_ids = {tool_call["id"]: tool_call for tool_call in ai_message.tool_calls}
    parts = []
    for i, message in enumerate(tool_messages):
        if not tool_calls_ids:
            break
        if message.tool_call_id in tool_calls_ids:
            tool_call = tool_calls_ids[message.tool_call_id]
            part = _convert_tool_message_to_part(message, name=tool_call.get("name"))
            parts.append(part)
            # remove the id from the dict, so that we do not iterate over it again
            tool_calls_ids.pop(message.tool_call_id)
    return parts


def _parse_chat_history(
    input_messages: Sequence[BaseMessage], convert_system_message_to_human: bool = False
) -> Tuple[Content | None, List[Content]]:
    messages: List[Content] = []

    if convert_system_message_to_human:
        warnings.warn("Convert_system_message_to_human will be deprecated!")

    system_instruction: Content | None = None
    messages_without_tool_messages = [
        message for message in input_messages if not isinstance(message, ToolMessage)
    ]
    tool_messages = [
        message for message in input_messages if isinstance(message, ToolMessage)
    ]
    for i, message in enumerate(messages_without_tool_messages):
        if isinstance(message, SystemMessage):
            system_parts = _convert_to_parts(message.content)
            if i == 0:
                system_instruction = Content(parts=system_parts)
            elif system_instruction is not None:
                system_instruction.parts.extend(system_parts)
            else:
                pass
            continue
        elif isinstance(message, AIMessage):
            role = "model"
            if message.tool_calls:
                ai_message_parts = []
                for tool_call in message.tool_calls:
                    function_call = FunctionCall(
                        name=tool_call["name"],
                        args=tool_call["args"],
                    )
                    ai_message_parts.append(Part(function_call=function_call.model_dump()))
                tool_messages_parts = _get_ai_message_tool_messages_parts(
                    tool_messages=tool_messages, ai_message=message
                )
                messages.append(Content(role=role, parts=ai_message_parts))
                messages.append(Content(role="user", parts=tool_messages_parts))
                continue
            elif raw_function_call := message.additional_kwargs.get("function_call"):
                function_call = FunctionCall(
                    name=raw_function_call["name"],
                    args=json.loads(raw_function_call["arguments"]),
                )
                parts = [Part(function_call=function_call.model_dump())]
            else:
                parts = _convert_to_parts(message.content)
        elif isinstance(message, HumanMessage):
            role = "user"
            parts = _convert_to_parts(message.content)
            if i == 1 and convert_system_message_to_human and system_instruction:
                parts = [p for p in system_instruction.parts] + parts
                system_instruction = None
        elif isinstance(message, FunctionMessage):
            role = "user"
            parts = [_convert_tool_message_to_part(message)]
        else:
            raise ValueError(
                f"Unexpected message with type {type(message)} at the position {i}."
            )

        messages.append(Content(role=role, parts=parts))
    return system_instruction, messages


def _parse_response_candidate(
    response_candidate: Dict[str, Any], streaming: bool = False
) -> AIMessage:
    content: Union[None, str, List[Union[str, dict]]] = None
    additional_kwargs = {}
    tool_calls = []
    invalid_tool_calls = []
    tool_call_chunks = []

    for part in response_candidate.get("content", {}).get("parts", []):
        try:
            text: str | None = part.get("text")
            # Remove erroneous newline character if present
            if text is not None:
                text = text.rstrip("\n")
        except AttributeError:
            text = None

        if text is not None:
            if not content:
                content = text
            elif isinstance(content, str) and text:
                content = [content, text]
            elif isinstance(content, list) and text:
                content.append(text)
            elif text:
                raise Exception("Unexpected content type")

        if part.get("inlineData", {}).get("mimeType", "").startswith("image/"):
            image_format = part["inlineData"]["mimeType"][6:]
            message = {
                "type": "image_url",
                "image_url": {
                    "url": image_bytes_to_b64_string(
                        part["inlineData"]["data"], image_format=image_format
                    )
                },
            }

            if not content:
                content = [message]
            elif isinstance(content, str) and message:
                content = [content, message]
            elif isinstance(content, list) and message:
                content.append(message)
            elif message:
                raise Exception("Unexpected content type")

        if part.get("functionCall"):
            function_call = {"name": part["functionCall"]["name"]}
            function_call["arguments"] = json.dumps(part["functionCall"].get("args", {}))
            additional_kwargs["function_call"] = function_call

            if streaming:
                tool_call_chunks.append(
                    tool_call_chunk(
                        name=function_call.get("name"),
                        args=function_call.get("arguments"),
                        id=function_call.get("id", str(uuid.uuid4())),
                        index=function_call.get("index"),  # type: ignore
                    )
                )
            else:
                try:
                    tool_call_dict = parse_tool_calls(
                        [{"function": function_call}],
                        return_id=False,
                    )[0]
                except Exception as e:
                    invalid_tool_calls.append(
                        invalid_tool_call(
                            name=function_call.get("name"),
                            args=function_call.get("arguments"),
                            id=function_call.get("id", str(uuid.uuid4())),
                            error=str(e),
                        )
                    )
                else:
                    tool_calls.append(
                        tool_call(
                            name=tool_call_dict["name"],
                            args=tool_call_dict["args"],
                            id=tool_call_dict.get("id", str(uuid.uuid4())),
                        )
                    )
    if content is None:
        content = ""

    if streaming:
        return AIMessageChunk(
            content=cast(Union[str, List[Union[str, Dict[Any, Any]]]], content),
            additional_kwargs=additional_kwargs,
            tool_call_chunks=tool_call_chunks,
        )

    return AIMessage(
        content=cast(Union[str, List[Union[str, Dict[Any, Any]]]], content),
        additional_kwargs=additional_kwargs,
        tool_calls=tool_calls,
        invalid_tool_calls=invalid_tool_calls,
    )


def _response_to_result(
    response: Dict[str, Any],
    stream: bool = False,
    prev_usage: UsageMetadata | None = None,
) -> ChatResult:
    """Converts a Gemini API response into a LangChain ChatResult."""
    llm_output = {"prompt_feedback": response.get("promptFeedback", {})}

    # previous usage metadata needs to be subtracted because gemini api returns
    # already-accumulated token counts with each chunk
    prev_input_tokens = prev_usage["input_tokens"] if prev_usage else 0
    prev_output_tokens = prev_usage["output_tokens"] if prev_usage else 0
    prev_total_tokens = prev_usage["total_tokens"] if prev_usage else 0

    # Get usage metadata
    try:
        usage_metadata = response.get("usageMetadata", {})
        input_tokens = usage_metadata.get("promptTokenCount", 0)
        output_tokens = usage_metadata.get("candidatesTokenCount", 0)
        total_tokens = usage_metadata.get("totalTokenCount", 0)
        cache_read_tokens = usage_metadata.get("cachedContentTokenCount", 0)
        if input_tokens + output_tokens + cache_read_tokens + total_tokens > 0:
            lc_usage = UsageMetadata(
                input_tokens=input_tokens - prev_input_tokens,
                output_tokens=output_tokens - prev_output_tokens,
                total_tokens=total_tokens - prev_total_tokens,
                input_token_details={"cache_read": cache_read_tokens},
            )
        else:
            lc_usage = None
    except AttributeError:
        lc_usage = None

    generations: List[ChatGeneration] = []

    for candidate in response.get("candidates", []):
        generation_info = {}
        if candidate.get("finishReason"):
            generation_info["finish_reason"] = candidate["finishReason"]
        generation_info["safety_ratings"] = candidate.get("safetyRatings", [])
        message = _parse_response_candidate(candidate, streaming=stream)
        message.usage_metadata = lc_usage
        if stream:
            generations.append(
                ChatGenerationChunk(
                    message=cast(AIMessageChunk, message),
                    generation_info=generation_info,
                )
            )
        else:
            generations.append(
                ChatGeneration(message=message, generation_info=generation_info)
            )
    if not response.get("candidates"):
        # Likely a "prompt feedback" violation (e.g., toxic input)
        # Raising an error would be different than how OpenAI handles it,
        # so we'll just log a warning and continue with an empty message.
        logger.warning(
            "Gemini produced an empty response. Continuing with empty message\n"
            f"Feedback: {response.get('promptFeedback')}"
        )
        if stream:
            generations = [
                ChatGenerationChunk(
                    message=AIMessageChunk(content=""), generation_info={}
                )
            ]
        else:
            generations = [ChatGeneration(message=AIMessage(""), generation_info={})]
    return ChatResult(generations=generations, llm_output=llm_output)


def _is_event_loop_running() -> bool:
    try:
        asyncio.get_running_loop()
        return True
    except RuntimeError:
        return False


class GeminiRestClient:
    """A REST client for making requests to the Gemini API."""

    def __init__(
        self,
        api_key: str,
        base_url: str,
        headers: Dict[str, str] | None = None,
        timeout: float | None = None,
    ):
        self.api_key = api_key
        self.base_url = base_url.rstrip("/")
        self.headers = headers or {}
        self.timeout = timeout
        self._client = None
        self._async_client = None

    def _get_client(self) -> httpx.Client:
        if self._client is None:
            self._client = httpx.Client(
                base_url=self.base_url,
                headers={
                    "Content-Type": "application/json",
                    **self.headers,
                },
                timeout=self.timeout,
            )
        return self._client

    async def _get_async_client(self) -> httpx.AsyncClient:
        if self._async_client is None:
            self._async_client = httpx.AsyncClient(
                base_url=self.base_url,
                headers={
                    "Content-Type": "application/json",
                    **self.headers,
                },
                timeout=self.timeout,
            )
        return self._async_client

    def _convert_to_dict(self, obj: Any) -> Any:
        """Convert Pydantic models and other objects to dictionaries."""
        if hasattr(obj, 'model_dump'):
            return obj.model_dump()
        elif isinstance(obj, list):
            return [self._convert_to_dict(item) for item in obj]
        elif isinstance(obj, dict):
            return {k: self._convert_to_dict(v) for k, v in obj.items()}
        return obj

    def _prepare_payload(
        self,
        contents: List[Dict[str, Any]],
        generation_config: Dict[str, Any] | None = None,
        safety_settings: List[Dict[str, Any]] | None = None,
        tools: List[Dict[str, Any]] | None = None,
        tool_config: Dict[str, Any] | None = None,
        system_instruction: Dict[str, Any] | None = None,
        cached_content: str | None = None,
    ) -> Dict[str, Any]:
        """Prepare the payload for Gemini API requests."""
        payload = {
            "contents": self._convert_to_dict(contents),
            "generationConfig": self._convert_to_dict(generation_config or {}),
            "safetySettings": self._convert_to_dict(safety_settings or []),
        }
        if tools:
            payload["tools"] = self._convert_to_dict(tools)
        if tool_config:
            payload["toolConfig"] = self._convert_to_dict(tool_config)
        if system_instruction:
            payload["systemInstruction"] = self._convert_to_dict(system_instruction)
        if cached_content:
            payload["cachedContent"] = cached_content
        return payload

    def generate_content(
        self,
        model: str,
        contents: List[Dict[str, Any]],
        generation_config: Dict[str, Any] | None = None,
        safety_settings: List[Dict[str, Any]] | None = None,
        tools: List[Dict[str, Any]] | None = None,
        tool_config: Dict[str, Any] | None = None,
        system_instruction: Dict[str, Any] | None = None,
        cached_content: str | None = None,
    ) -> Dict[str, Any]:
        """Generate content using the Gemini API."""
        payload = self._prepare_payload(
            contents=contents,
            generation_config=generation_config,
            safety_settings=safety_settings,
            tools=tools,
            tool_config=tool_config,
            system_instruction=system_instruction,
            cached_content=cached_content,
        )

        response = self._get_client().post(
            f"v1beta/{model}:generateContent",
            json=payload,
        )
        response.raise_for_status()
        return response.json()

    async def generate_content_async(
        self,
        model: str,
        contents: List[Dict[str, Any]],
        generation_config: Dict[str, Any] | None = None,
        safety_settings: List[Dict[str, Any]] | None = None,
        tools: List[Dict[str, Any]] | None = None,
        tool_config: Dict[str, Any] | None = None,
        system_instruction: Dict[str, Any] | None = None,
        cached_content: str | None = None,
    ) -> Dict[str, Any]:
        """Generate content asynchronously using the Gemini API."""
        payload = self._prepare_payload(
            contents=contents,
            generation_config=generation_config,
            safety_settings=safety_settings,
            tools=tools,
            tool_config=tool_config,
            system_instruction=system_instruction,
            cached_content=cached_content,
        )
        client = await self._get_async_client()
        response = await client.post(
            f"v1beta/{model}:generateContent",
            json=payload,
        )
        response.raise_for_status()
        return response.json()

    def stream_generate_content(
        self,
        model: str,
        contents: List[Dict[str, Any]],
        generation_config: Dict[str, Any] | None = None,
        safety_settings: List[Dict[str, Any]] | None = None,
        tools: List[Dict[str, Any]] | None = None,
        tool_config: Dict[str, Any] | None = None,
        system_instruction: Dict[str, Any] | None = None,
        cached_content: str | None = None,
    ) -> Iterator[Dict[str, Any]]:
        """Stream content generation using the Gemini API."""
        payload = self._prepare_payload(
            contents=contents,
            generation_config=generation_config,
            safety_settings=safety_settings,
            tools=tools,
            tool_config=tool_config,
            system_instruction=system_instruction,
            cached_content=cached_content,
        )


        with self._get_client().stream(
            "POST",
            f"v1beta/{model}:streamGenerateContent",
            json=payload,
        ) as response:
            response.raise_for_status()
            for line in response.iter_lines():
                if line:
                    yield json.loads(line)

    async def stream_generate_content_async(
        self,
        model: str,
        contents: List[Dict[str, Any]],
        generation_config: Dict[str, Any] | None = None,
        safety_settings: List[Dict[str, Any]] | None = None,
        tools: List[Dict[str, Any]] | None = None,
        tool_config: Dict[str, Any] | None = None,
        system_instruction: Dict[str, Any] | None = None,
        cached_content: str | None = None,
    ) -> AsyncIterator[Dict[str, Any]]:
        """Stream content generation asynchronously using the Gemini API."""
        payload = self._prepare_payload(
            contents=contents,
            generation_config=generation_config,
            safety_settings=safety_settings,
            tools=tools,
            tool_config=tool_config,
            system_instruction=system_instruction,
            cached_content=cached_content,
        )

        client = await self._get_async_client()
        async with client.stream(
            "POST",
            f"v1beta/{model}:streamGenerateContent",
            json=payload,
        ) as response:
            response.raise_for_status()
            async for line in response.aiter_lines():
                if line:
                    yield json.loads(line)

    def count_tokens(
        self,
        model: str,
        contents: List[Dict[str, Any]],
    ) -> Dict[str, Any]:
        """Count tokens in the input using the Gemini API."""
        payload = {"contents": self._convert_to_dict(contents)}
        response = self._get_client().post(
            f"v1beta/{model}:countTokens",
            json=payload,
        )
        response.raise_for_status()
        return response.json()


class ChatGoogleGenerativeAI(BaseChatModel):
    """`Google AI` chat models integration."""

    client: GeminiRestClient | None = Field(default=None, exclude=True)
    async_client: GeminiRestClient | None = Field(default=None, exclude=True)
    default_metadata: Sequence[Tuple[str, str]] = Field(default_factory=list)
    convert_system_message_to_human: bool = False
    cached_content: str | None = None
    model: str = Field(default="gemini-pro")
    google_api_key: SecretStr | None = Field(default=None)
    temperature: float | None = Field(default=None)
    top_p: float | None = Field(default=None)
    top_k: int | None = Field(default=None)
    n: int = Field(default=1)
    max_output_tokens: int | None = Field(default=None)
    safety_settings: Dict[str, str] | None = Field(default=None)
    response_modalities: List[str] | None = Field(default=None)
    client_options: Dict[str, Any] | None = Field(default=None)
    additional_headers: Dict[str, str] | None = Field(default=None)
    transport: str | None = Field(default="rest")

    model_config = ConfigDict(
        populate_by_name=True,
    )

    @property
    def lc_secrets(self) -> Dict[str, str]:
        return {"google_api_key": "GOOGLE_API_KEY"}

    @property
    def _llm_type(self) -> str:
        return "chat-google-generative-ai"

    @classmethod
    def is_lc_serializable(self) -> bool:
        return True

    @model_validator(mode="after")
    def validate_environment(self) -> Self:
        """Validates params and initializes the REST client."""
        if self.temperature is not None and not 0 <= self.temperature <= 2.0:
            raise ValueError("temperature must be in the range [0.0, 2.0]")

        if self.top_p is not None and not 0 <= self.top_p <= 1:
            raise ValueError("top_p must be in the range [0.0, 1.0]")

        if self.top_k is not None and self.top_k <= 0:
            raise ValueError("top_k must be positive")

        additional_headers = self.additional_headers or {}
        self.default_metadata = tuple(additional_headers.items())

        google_api_key = None
        if isinstance(self.google_api_key, SecretStr):
            google_api_key = self.google_api_key.get_secret_value()
        else:
            google_api_key = self.google_api_key

        base_url = self.client_options.get("api_endpoint", "https://generativelanguage.googleapis.com")
        self.client = GeminiRestClient(
            api_key=google_api_key,
            base_url=base_url,
            headers=additional_headers,
        )
        self.async_client = GeminiRestClient(
            api_key=google_api_key,
            base_url=base_url,
            headers=additional_headers,
        )
        return self

    def _get_ls_params(
        self, stop: List[str] | None = None, **kwargs: Any
    ) -> LangSmithParams:
        """Get standard params for tracing."""
        params = self._get_invocation_params(stop=stop, **kwargs)
        ls_params = LangSmithParams(
            ls_provider="google_genai",
            ls_model_name=self.model,
            ls_model_type="chat",
            ls_temperature=params.get("temperature", self.temperature),
        )
        if ls_max_tokens := params.get("max_output_tokens", self.max_output_tokens):
            ls_params["ls_max_tokens"] = ls_max_tokens
        if ls_stop := stop or params.get("stop", None):
            ls_params["ls_stop"] = ls_stop
        return ls_params

    def _prepare_params(
        self,
        stop: List[str] | None,
        generation_config: Dict[str, Any] | None = None,
    ) -> GenerationConfig:
        gen_config = {
            k: v
            for k, v in {
                "candidate_count": self.n,
                "temperature": self.temperature,
                "stop_sequences": stop,
                "max_output_tokens": self.max_output_tokens,
                "top_k": self.top_k,
                "top_p": self.top_p,
                "response_modalities": self.response_modalities,
            }.items()
            if v is not None
        }
        if generation_config:
            gen_config = {**gen_config, **generation_config}
        return GenerationConfig(**gen_config)

    def _generate(
        self,
        messages: List[BaseMessage],
        stop: List[str] | None = None,
        run_manager: CallbackManagerForLLMRun | None = None,
        *,
        tools: Sequence[Union[_ToolDict, GoogleTool]] | None = None,
        functions: Sequence[_FunctionDeclarationType] | None = None,
        safety_settings: SafetySettingDict | None = None,
        tool_config: Union[Dict, _ToolConfigDict] | None = None,
        generation_config: Dict[str, Any] | None = None,
        cached_content: str | None = None,
        tool_choice: Union[_ToolChoiceType, bool] | None = None,
        **kwargs: Any,
    ) -> ChatResult:
        request = self._prepare_request(
            messages,
            stop=stop,
            tools=tools,
            functions=functions,
            safety_settings=safety_settings,
            tool_config=tool_config,
            generation_config=generation_config,
            cached_content=cached_content or self.cached_content,
            tool_choice=tool_choice,
        )
        response = _chat_with_retry(
            request=request,
            **kwargs,
            generation_method=self.client.generate_content,
            metadata=self.default_metadata,
        )
        return _response_to_result(response)

    async def _agenerate(
        self,
        messages: List[BaseMessage],
        stop: List[str] | None = None,
        run_manager: AsyncCallbackManagerForLLMRun | None = None,
        *,
        tools: Sequence[Union[_ToolDict, GoogleTool]] | None = None,
        functions: Sequence[_FunctionDeclarationType] | None = None,
        safety_settings: SafetySettingDict | None = None,
        tool_config: Union[Dict, _ToolConfigDict] | None = None,
        generation_config: Dict[str, Any] | None = None,
        cached_content: str | None = None,
        tool_choice: Union[_ToolChoiceType, bool] | None = None,
        **kwargs: Any,
    ) -> ChatResult:
        request = self._prepare_request(
            messages,
            stop=stop,
            tools=tools,
            functions=functions,
            safety_settings=safety_settings,
            tool_config=tool_config,
            generation_config=generation_config,
            cached_content=cached_content or self.cached_content,
            tool_choice=tool_choice,
        )
        response = await _achat_with_retry(
            request=request,
            **kwargs,
            generation_method=self.async_client.generate_content_async,
            metadata=self.default_metadata,
        )
        return _response_to_result(response)

    def _stream(
        self,
        messages: List[BaseMessage],
        stop: List[str] | None = None,
        run_manager: CallbackManagerForLLMRun | None = None,
        *,
        tools: Sequence[Union[_ToolDict, GoogleTool]] | None = None,
        functions: Sequence[_FunctionDeclarationType] | None = None,
        safety_settings: SafetySettingDict | None = None,
        tool_config: Union[Dict, _ToolConfigDict] | None = None,
        generation_config: Dict[str, Any] | None = None,
        cached_content: str | None = None,
        tool_choice: Union[_ToolChoiceType, bool] | None = None,
        **kwargs: Any,
    ) -> Iterator[ChatGenerationChunk]:
        request = self._prepare_request(
            messages,
            stop=stop,
            tools=tools,
            functions=functions,
            safety_settings=safety_settings,
            tool_config=tool_config,
            generation_config=generation_config,
            cached_content=cached_content or self.cached_content,
            tool_choice=tool_choice,
        )
        response = _chat_with_retry(
            request=request,
            generation_method=self.client.stream_generate_content,
            **kwargs,
            metadata=self.default_metadata,
        )

        prev_usage_metadata: UsageMetadata | None = None
        for chunk in response:
            _chat_result = _response_to_result(
                chunk, stream=True, prev_usage=prev_usage_metadata
            )
            gen = cast(ChatGenerationChunk, _chat_result.generations[0])
            message = cast(AIMessageChunk, gen.message)

            curr_usage_metadata: UsageMetadata | dict[str, int] = (
                message.usage_metadata or {}
            )

            prev_usage_metadata = (
                message.usage_metadata
                if prev_usage_metadata is None
                else UsageMetadata(
                    input_tokens=prev_usage_metadata.get("input_tokens", 0)
                    + curr_usage_metadata.get("input_tokens", 0),
                    output_tokens=prev_usage_metadata.get("output_tokens", 0)
                    + curr_usage_metadata.get("output_tokens", 0),
                    total_tokens=prev_usage_metadata.get("total_tokens", 0)
                    + curr_usage_metadata.get("total_tokens", 0),
                )
            )

            if run_manager:
                run_manager.on_llm_new_token(gen.text)
            yield gen

    async def _astream(
        self,
        messages: List[BaseMessage],
        stop: List[str] | None = None,
        run_manager: AsyncCallbackManagerForLLMRun | None = None,
        *,
        tools: Sequence[Union[_ToolDict, GoogleTool]] | None = None,
        functions: Sequence[_FunctionDeclarationType] | None = None,
        safety_settings: SafetySettingDict | None = None,
        tool_config: Union[Dict, _ToolConfigDict] | None = None,
        generation_config: Dict[str, Any] | None = None,
        cached_content: str | None = None,
        tool_choice: Union[_ToolChoiceType, bool] | None = None,
        **kwargs: Any,
    ) -> AsyncIterator[ChatGenerationChunk]:
        request = self._prepare_request(
            messages,
            stop=stop,
            tools=tools,
            functions=functions,
            safety_settings=safety_settings,
            tool_config=tool_config,
            generation_config=generation_config,
            cached_content=cached_content or self.cached_content,
            tool_choice=tool_choice,
        )
        prev_usage_metadata: UsageMetadata | None = None
        async for chunk in await _achat_with_retry(
            request=request,
            generation_method=self.async_client.stream_generate_content_async,
            **kwargs,
            metadata=self.default_metadata,
        ):
            _chat_result = _response_to_result(
                chunk, stream=True, prev_usage=prev_usage_metadata
            )
            gen = cast(ChatGenerationChunk, _chat_result.generations[0])
            message = cast(AIMessageChunk, gen.message)

            curr_usage_metadata: UsageMetadata | dict[str, int] = (
                message.usage_metadata or {}
            )

            prev_usage_metadata = (
                message.usage_metadata
                if prev_usage_metadata is None
                else UsageMetadata(
                    input_tokens=prev_usage_metadata.get("input_tokens", 0)
                    + curr_usage_metadata.get("input_tokens", 0),
                    output_tokens=prev_usage_metadata.get("output_tokens", 0)
                    + curr_usage_metadata.get("output_tokens", 0),
                    total_tokens=prev_usage_metadata.get("total_tokens", 0)
                    + curr_usage_metadata.get("total_tokens", 0),
                )
            )

            if run_manager:
                await run_manager.on_llm_new_token(gen.text)
            yield gen

    def _prepare_request(
        self,
        messages: List[BaseMessage],
        *,
        stop: List[str] | None = None,
        tools: Sequence[Union[_ToolDict, GoogleTool]] | None = None,
        functions: Sequence[_FunctionDeclarationType] | None = None,
        safety_settings: SafetySettingDict | None = None,
        tool_config: Union[Dict, _ToolConfigDict] | None = None,
        tool_choice: Union[_ToolChoiceType, bool] | None = None,
        generation_config: Dict[str, Any] | None = None,
        cached_content: str | None = None,
    ) -> Dict[str, Any]:
        if tool_choice and tool_config:
            raise ValueError(
                "Must specify at most one of tool_choice and tool_config, received "
                f"both:\n\n{tool_choice=}\n\n{tool_config=}"
            )
        formatted_tools = None
        if tools:
            formatted_tools = [convert_to_genai_function_declarations(tools)]
        elif functions:
            formatted_tools = [convert_to_genai_function_declarations(functions)]

        filtered_messages = []
        for message in messages:
            if isinstance(message, HumanMessage) and not message.content:
                warnings.warn(
                    "HumanMessage with empty content was removed to prevent API error"
                )
            else:
                filtered_messages.append(message)
        messages = filtered_messages

        system_instruction, history = _parse_chat_history(
            messages,
            convert_system_message_to_human=self.convert_system_message_to_human,
        )
        if tool_choice:
            if not formatted_tools:
                msg = (
                    f"Received {tool_choice=} but no {tools=}. 'tool_choice' can only "
                    f"be specified if 'tools' is specified."
                )
                raise ValueError(msg)
            all_names = [
                f.name for t in formatted_tools for f in t.function_declarations
            ]
            tool_config = _tool_choice_to_tool_config(tool_choice, all_names)

        formatted_tool_config = None
        if tool_config:
            formatted_tool_config = ToolConfig(
                function_calling_config=tool_config["function_calling_config"]
            )
        formatted_safety_settings = []
        if safety_settings:
            formatted_safety_settings = [
                SafetySetting(category=c, threshold=t)
                for c, t in safety_settings.items()
            ]

        # Construct the full model path
        model_path = f"models/{self.model}"

        request = {
            "model": model_path,
            "contents": history,
            "tools": formatted_tools,
            "tool_config": formatted_tool_config,
            "safety_settings": formatted_safety_settings,
            "generation_config": self._prepare_params(
                stop, generation_config=generation_config
            ),
            "cached_content": cached_content,
        }
        if system_instruction:
            request["system_instruction"] = system_instruction

        return request

    def get_num_tokens(self, text: str) -> int:
        """Get the number of tokens present in the text.

        Useful for checking if an input will fit in a model's context window.

        Args:
            text: The string input to tokenize.

        Returns:
            The integer number of tokens in the text.
        """
        result = self.client.count_tokens(
            model=self.model,
            contents=[{"parts": [{"text": text}]}],
        )
        return result["total_tokens"]

    def with_structured_output(
        self,
        schema: Union[Dict, Type[BaseModel]],
        *,
        include_raw: bool = False,
        **kwargs: Any,
    ) -> Runnable[LanguageModelInput, Union[Dict, BaseModel]]:
        _ = kwargs.pop("method", None)
        _ = kwargs.pop("strict", None)
        if kwargs:
            raise ValueError(f"Received unsupported arguments {kwargs}")
        tool_name = _get_tool_name(schema)  # type: ignore[arg-type]
        if isinstance(schema, type) and is_basemodel_subclass_safe(schema):
            parser: OutputParserLike = PydanticToolsParser(
                tools=[schema], first_tool_only=True
            )
        else:
            global WARNED_STRUCTURED_OUTPUT_JSON_MODE
            warnings.warn(
                "ChatGoogleGenerativeAI.with_structured_output with dict schema has "
                "changed recently to align with behavior of other LangChain chat "
                "models. More context: "
                "https://github.com/langchain-ai/langchain-google/pull/772"
            )
            WARNED_STRUCTURED_OUTPUT_JSON_MODE = True
            parser = JsonOutputKeyToolsParser(key_name=tool_name, first_tool_only=True)
        tool_choice = tool_name if self._supports_tool_choice else None
        try:
            llm = self.bind_tools(
                [schema],
                tool_choice=tool_choice,
                ls_structured_output_format={
                    "kwargs": {"method": "function_calling"},
                    "schema": convert_to_openai_tool(schema),
                },
            )
        except Exception:
            llm = self.bind_tools([schema], tool_choice=tool_choice)
        if include_raw:
            parser_with_fallback = RunnablePassthrough.assign(
                parsed=itemgetter("raw") | parser, parsing_error=lambda _: None
            ).with_fallbacks(
                [RunnablePassthrough.assign(parsed=lambda _: None)],
                exception_key="parsing_error",
            )
            return {"raw": llm} | parser_with_fallback
        else:
            return llm | parser

    def bind_tools(
        self,
        tools: Sequence[
            dict[str, Any] | type | Callable[..., Any] | BaseTool | GoogleTool
        ],
        tool_config: Union[Dict, _ToolConfigDict] | None = None,
        *,
        tool_choice: Union[_ToolChoiceType, bool] | None = None,
        **kwargs: Any,
    ) -> Runnable[LanguageModelInput, BaseMessage]:
        """Bind tool-like objects to this chat model.

        Assumes model is compatible with google-generativeAI tool-calling API.

        Args:
            tools: A list of tool definitions to bind to this chat model.
                Can be a pydantic model, callable, or BaseTool. Pydantic
                models, callables, and BaseTools will be automatically converted to
                their schema dictionary representation.
            **kwargs: Any additional parameters to pass to the
                :class:`~langchain.runnable.Runnable` constructor.
        """
        if tool_choice and tool_config:
            raise ValueError(
                "Must specify at most one of tool_choice and tool_config, received "
                f"both:\n\n{tool_choice=}\n\n{tool_config=}"
            )
        try:
            formatted_tools: list = [convert_to_openai_tool(tool) for tool in tools]  # type: ignore[arg-type]
        except Exception:
            formatted_tools = [
                tool_to_dict(convert_to_genai_function_declarations(tools))
            ]
        if tool_choice:
            kwargs["tool_choice"] = tool_choice
        elif tool_config:
            kwargs["tool_config"] = tool_config
        else:
            pass
        return self.bind(tools=formatted_tools, **kwargs)

    @property
    def _supports_tool_choice(self) -> bool:
        return (
            "gemini-1.5-pro" in self.model
            or "gemini-1.5-flash" in self.model
            or "gemini-2" in self.model
        )


def _get_tool_name(
    tool: Union[_ToolDict, GoogleTool, Dict],
) -> str:
    try:
        genai_tool = tool_to_dict(convert_to_genai_function_declarations([tool]))
        return [f["name"] for f in genai_tool["function_declarations"]][0]  # type: ignore[index]
    except ValueError as e:  # other TypedDict
        if is_typeddict(tool):
            return convert_to_openai_tool(cast(Dict, tool))["function"]["name"]
        else:
            raise e

def _tool_choice_to_tool_config(
    tool_choice: Union[str, bool, Dict[str, Any]], all_names: List[str]
) -> Dict[str, Any]:
    """Convert tool_choice to tool_config format."""
    if isinstance(tool_choice, bool):
        return {
            "function_calling_config": {
                "mode": "AUTO" if tool_choice else "NONE",
            }
        }
    elif isinstance(tool_choice, str):
        if tool_choice not in all_names:
            raise ValueError(
                f"Tool choice {tool_choice} not found in available tools: {all_names}"
            )
        return {
            "function_calling_config": {
                "mode": "ANY",
                "allowed_function_names": [tool_choice],
            }
        }
    else:
        return tool_choice

def convert_to_genai_function_declarations(
    tools: Sequence[Union[Dict[str, Any], Type[BaseModel], Callable[..., Any], BaseTool]]
) -> Dict[str, Any]:
    """Convert tools to Gemini function declarations format."""
    function_declarations = []
    for tool in tools:
        if isinstance(tool, dict):
            fn = tool.get("function", {})
            fn_parameters = fn.get("parameters", {})
            function_declarations.append({
                "name": fn.get("name", ""),
                "description": fn.get("description", ""),
                "parameters": {
                    "type": "object",
                    "properties": fn_parameters.get("properties", {}),
                    "required": fn_parameters.get("required", []),
                },
            })
        elif isinstance(tool, type) and issubclass(tool, BaseModel):
            schema = tool.model_json_schema()
            function_declarations.append({
                "name": schema.get("title", ""),
                "description": schema.get("description", ""),
                "parameters": {
                    "type": "object",
                    "properties": {
                        name: {
                            "type": prop.get("type", "string"),
                            "description": prop.get("description", ""),
                        }
                        for name, prop in schema.get("properties", {}).items()
                    },
                    "required": schema.get("required", []),
                },
            })
        elif callable(tool):
            # For callables, we'll create a basic function declaration
            function_declarations.append({
                "name": tool.__name__,
                "description": tool.__doc__ or "",
                "parameters": {
                    "type": "object",
                    "properties": {},
                    "required": [],
                },
            })
        elif isinstance(tool, BaseTool):
            function_declarations.append({
                "name": tool.name,
                "description": tool.description,
                "parameters": {
                    "type": "object",
                    "properties": {},
                    "required": [],
                },
            })
    return {"functionDeclarations": function_declarations}

def is_basemodel_subclass_safe(cls: Type[Any]) -> bool:
    """Check if a class is a safe subclass of BaseModel."""
    try:
        return issubclass(cls, BaseModel)
    except TypeError:
        return False

def tool_to_dict(tool: Dict[str, Any]) -> Dict[str, Any]:
    """Convert a tool to a dictionary format."""
    return tool

def image_bytes_to_b64_string(image_bytes: bytes, image_format: str = "jpeg") -> str:
    """Convert image bytes to base64 string."""
    import base64
    return f"data:image/{image_format};base64,{base64.b64encode(image_bytes).decode('utf-8')}"



import logging
import queue
import threading

from opentelemetry.context import attach, create_key, detach, set_value
from opentelemetry.sdk._logs import LogData, LogRecordProcessor
from opentelemetry.sdk._logs._internal.export import LogExporter

_logger = logging.getLogger(__name__)

class AsyncLogRecordProcessor(LogRecordProcessor):
    """This is an implementation of LogRecordProcessor which passes
    received logs in the export-friendly LogData representation to the
    configured LogExporter asynchronously using a background thread.
    """

    def __init__(self, exporter: LogExporter):
        self._exporter = exporter
        self._shutdown = False
        self._queue: queue.Queue = queue.Queue()
        self._worker_thread: threading.Thread | None = None
        self._start_worker()

    def _start_worker(self):
        """Start the background worker thread."""
        self._worker_thread = threading.Thread(
            target=self._process_logs,
            daemon=True,
            name="LogExporterWorker"
        )
        self._worker_thread.start()

    def _process_logs(self):
        """Process logs from the queue in the background thread."""
        while not self._shutdown:
            try:
                log_data = self._queue.get(timeout=1.0)
                token = attach(set_value(create_key("suppress_instrumentation"), True))
                try:
                    self._exporter.export((log_data,))
                except Exception:  # pylint: disable=broad-exception-caught
                    _logger.exception("Exception while exporting logs.")
                finally:
                    detach(token)
                    self._queue.task_done()
            except queue.Empty:
                continue

    def emit(self, log_data: LogData):
        if self._shutdown:
            return
        self._queue.put(log_data)

    def shutdown(self):
        """Shutdown the processor and wait for pending exports to complete."""
        self._shutdown = True
        if self._worker_thread:
            self._worker_thread.join(timeout=5.0)
        self._exporter.shutdown()

    def force_flush(self, timeout_millis: int = 500) -> bool:
        """Wait for all pending exports to complete."""
        try:
            self._queue.join(timeout=timeout_millis)
            return True
        except Exception:  # pylint: disable=broad-exception-caught
            return False



import logging

from ..client import client
from ..client.response_interceptor import response_interceptors_async, response_interceptors_sync
from ..sandbox.client import client as client_sandbox
from .settings import settings

logger = logging.getLogger(__name__)


def telemetry() -> None:
    from blaxel.telemetry import telemetry_manager

    telemetry_manager.initialize(settings)


def autoload() -> None:
    client.with_base_url(settings.base_url)
    client.with_auth(settings.auth)

    # Register response interceptors for authentication error handling
    # Access the underlying httpx clients and add event hooks
    # Use sync interceptors for sync clients and async interceptors for async clients
    httpx_client = client.get_httpx_client()
    httpx_client.event_hooks["response"] = response_interceptors_sync

    httpx_async_client = client.get_async_httpx_client()
    httpx_async_client.event_hooks["response"] = response_interceptors_async

    httpx_sandbox_client = client_sandbox.get_httpx_client()
    httpx_sandbox_client.event_hooks["response"] = response_interceptors_sync

    httpx_sandbox_async_client = client_sandbox.get_async_httpx_client()
    httpx_sandbox_async_client.event_hooks["response"] = response_interceptors_async

    try:
        telemetry()
    except Exception:
        pass



import asyncio
import sys
from pathlib import Path

from blaxel.core.sandbox import SandboxInstance


async def main():
    """Test file read and download operations."""
    try:
        # Get the directory where this test file is located
        test_dir = Path(__file__).parent
        archive_path = test_dir / "archive.zip"
        downloaded_path = test_dir / "archive.downloaded.zip"

        # Ensure archive exists
        if not archive_path.exists():
            print(f"❌ Test archive not found at {archive_path}")
            sys.exit(1)

        sandbox = await SandboxInstance.create()
        print("✓ Sandbox created")

        # Test text file write
        await sandbox.fs.write("/tmp/test.txt", "Hello world")
        print("✓ Text file written")

        # Test binary file write from local path
        await sandbox.fs.write_binary("/tmp/archive.zip", str(archive_path))
        print("✓ Binary file written")

        # Test ls
        listing = await sandbox.fs.ls("/tmp")
        assert any(f.name == "test.txt" for f in listing.files), "test.txt should exist in /tmp"
        assert any(f.name == "archive.zip" for f in listing.files), "archive.zip should exist in /tmp"
        print("✓ Files listed correctly")

        # Test read text file
        text_content = await sandbox.fs.read("/tmp/test.txt")
        assert text_content == "Hello world", "Text content should match"
        print("✓ Text file read correctly")

        # Test read_binary
        binary_bytes = await sandbox.fs.read_binary("/tmp/archive.zip")
        assert isinstance(binary_bytes, bytes), "read_binary should return bytes"
        print("✓ Binary file read correctly")

        text_content_binary = await sandbox.fs.read_binary("/tmp/test.txt")
        assert isinstance(text_content_binary, bytes), "read_binary should return bytes"
        print("✓ Text file read as binary correctly")

        # Test download
        await sandbox.fs.download("/tmp/archive.zip", str(downloaded_path))
        print("✓ Binary file downloaded")

        # Compare file sizes
        original_size = archive_path.stat().st_size
        downloaded_size = downloaded_path.stat().st_size

        print(f"Original file size: {original_size} bytes")
        print(f"Downloaded file size: {downloaded_size} bytes")
        assert downloaded_size == original_size, "Downloaded file size should match original"
        print("✓ File sizes match")

        # Cleanup
        await SandboxInstance.delete(sandbox.metadata.name)
        print("✓ Sandbox deleted")

        # Clean up downloaded file
        if downloaded_path.exists():
            downloaded_path.unlink()
            print("✓ Downloaded file cleaned up")

        print("\n✅ All tests passed!")

    except Exception as e:
        print(f"\n❌ Test failed with error: {e}")
        import traceback
        traceback.print_exc()
        sys.exit(1)


if __name__ == "__main__":
    asyncio.run(main())




import sys
import time
from typing import List

from blaxel.core import SyncCodeInterpreter


def main():
    print("🚀 [sync-interpreter] starting")

    interp = None
    try:
        print("🔧 [sync-interpreter] creating interpreter sandbox (jupyter-server)...")
        t0 = time.perf_counter()
        interp = SyncCodeInterpreter.create()
        # interp = SyncCodeInterpreter.get("sandbox-b40e7115")
        # interp = SyncCodeInterpreter(sandbox=Sandbox(metadata=Metadata(name="test")), force_url="http://localhost:8888")
        name = getattr(getattr(interp, "metadata", None), "name", None)
        print(f"✅ created: {name}")
        print(f"⏱️ create: {int((time.perf_counter() - t0) * 1000)} ms")


        # Try creating a context (skip if endpoint not available)
        try:
            print("🔧 [sync-interpreter] creating code context (python)...")
            t0 = time.perf_counter()
            ctx = interp.create_code_context(language="python")
            print(f"✅ context created: id={ctx.id}")
            print(f"⏱️ create_context: {int((time.perf_counter() - t0) * 1000)} ms")
        except Exception as e:
            print(f"⚠️ [sync-interpreter] create_code_context skipped: {e}")

        # Try running simple code (skip if endpoint not available)
        try:
            print("🔧 [sync-interpreter] running code...")
            stdout_lines: List[str] = []
            stderr_lines: List[str] = []
            results: List[object] = []
            errors: List[object] = []

            def on_stdout(msg):
                text = getattr(msg, "text", str(msg))
                stdout_lines.append(text)
                print(f"[stdout] {text}")

            def on_stderr(msg):
                text = getattr(msg, "text", str(msg))
                stderr_lines.append(text)
                print(f"[stderr] {text}")

            def on_result(res):
                results.append(res)
                print(f"[result] {res}")

            def on_error(err):
                errors.append(err)
                print(f"[error] {err}")

            t0 = time.perf_counter()
            interp.run_code(
                "print('Hello from interpreter')",
                language="python",
                on_stdout=on_stdout,
                on_stderr=on_stderr,
                on_result=on_result,
                on_error=on_error,
                timeout=30.0,
            )
            print(f"⏱️ run_code(hello): {int((time.perf_counter() - t0) * 1000)} ms")
            print(
                f"✅ run_code finished: stdout={len(stdout_lines)} stderr={len(stderr_lines)} "
                f"results={len(results)} errors={len(errors)}"
            )

            # Define a function in one run, then call it in another run (using a context)
            print("🔧 [sync-interpreter] define function in first run_code, call in second")
            try:
                stdout_lines.clear()
                stderr_lines.clear()
                results.clear()
                errors.clear()

                # First run: define a function
                t0 = time.perf_counter()
                interp.run_code(
                    "def add(a, b):\n    return a + b",
                    on_stdout=on_stdout,
                    on_stderr=on_stderr,
                    on_result=on_result,
                    on_error=on_error,
                    timeout=30.0,
                )
                print(f"⏱️ run_code(define): {int((time.perf_counter() - t0) * 1000)} ms")

                # Second run: call the function
                stdout_lines.clear()
                stderr_lines.clear()
                results.clear()
                errors.clear()

                t0 = time.perf_counter()
                interp.run_code(
                    "print(add(2, 3))",
                    on_stdout=on_stdout,
                    on_stderr=on_stderr,
                    on_result=on_result,
                    on_error=on_error,
                    timeout=30.0,
                )
                print(f"⏱️ run_code(call): {int((time.perf_counter() - t0) * 1000)} ms")

                # Expect to see "5" in stdout
                got_stdout = "".join(stdout_lines)
                if "5" not in got_stdout:
                    raise AssertionError(f"Expected function output '5', got stdout={got_stdout!r}")
                print("✅ function persisted across runs via context")
            except Exception as e2:
                print(f"⚠️ [sync-interpreter] two-step run_code skipped: {e2}")
        except Exception as e:
            print(f"⚠️ [sync-interpreter] run_code skipped: {e}")

        print("🎉 [sync-interpreter] done")
        sys.exit(0)

    except AssertionError as e:
        print(f"❌ [sync-interpreter] assertion failed: {e}")
        sys.exit(1)
    except Exception as e:
        print(f"❌ [sync-interpreter] error: {e}")
        sys.exit(1)
    finally:
        if interp:
            try:
                n = getattr(getattr(interp, "metadata", None), "name", None)
                if n:
                    SyncCodeInterpreter.delete(n)
                    print(f"🧹 [sync-interpreter] deleted {n}")
            except Exception as e:
                print(f"⚠️ [sync-interpreter] cleanup failed: {e}")


if __name__ == "__main__":
    main()





from http import HTTPStatus
from typing import Any, Union

import httpx

from ... import errors
from ...client import Client
from ...models.content_search_response import ContentSearchResponse
from ...models.error_response import ErrorResponse
from ...types import UNSET, Response, Unset


def _get_kwargs(
    path: str,
    *,
    query: str,
    case_sensitive: Union[Unset, bool] = UNSET,
    context_lines: Union[Unset, int] = UNSET,
    max_results: Union[Unset, int] = UNSET,
    file_pattern: Union[Unset, str] = UNSET,
    exclude_dirs: Union[Unset, str] = UNSET,
) -> dict[str, Any]:
    params: dict[str, Any] = {}

    params["query"] = query

    params["caseSensitive"] = case_sensitive

    params["contextLines"] = context_lines

    params["maxResults"] = max_results

    params["filePattern"] = file_pattern

    params["excludeDirs"] = exclude_dirs

    params = {k: v for k, v in params.items() if v is not UNSET and v is not None}

    _kwargs: dict[str, Any] = {
        "method": "get",
        "url": f"/filesystem-content-search/{path}",
        "params": params,
    }

    return _kwargs


def _parse_response(*, client: Client, response: httpx.Response) -> Union[ContentSearchResponse, ErrorResponse] | None:
    if response.status_code == 200:
        response_200 = ContentSearchResponse.from_dict(response.json())

        return response_200
    if response.status_code == 400:
        response_400 = ErrorResponse.from_dict(response.json())

        return response_400
    if response.status_code == 422:
        response_422 = ErrorResponse.from_dict(response.json())

        return response_422
    if response.status_code == 500:
        response_500 = ErrorResponse.from_dict(response.json())

        return response_500
    if client.raise_on_unexpected_status:
        raise errors.UnexpectedStatus(response.status_code, response.content)
    else:
        return None


def _build_response(
    *, client: Client, response: httpx.Response
) -> Response[Union[ContentSearchResponse, ErrorResponse]]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(client=client, response=response),
    )


def sync_detailed(
    path: str,
    *,
    client: Union[Client],
    query: str,
    case_sensitive: Union[Unset, bool] = UNSET,
    context_lines: Union[Unset, int] = UNSET,
    max_results: Union[Unset, int] = UNSET,
    file_pattern: Union[Unset, str] = UNSET,
    exclude_dirs: Union[Unset, str] = UNSET,
) -> Response[Union[ContentSearchResponse, ErrorResponse]]:
    """Search for text content in files

     Searches for text content inside files using ripgrep. Returns matching lines with context.

    Args:
        path (str):
        query (str):
        case_sensitive (Union[Unset, bool]):
        context_lines (Union[Unset, int]):
        max_results (Union[Unset, int]):
        file_pattern (Union[Unset, str]):
        exclude_dirs (Union[Unset, str]):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[ContentSearchResponse, ErrorResponse]]
    """

    kwargs = _get_kwargs(
        path=path,
        query=query,
        case_sensitive=case_sensitive,
        context_lines=context_lines,
        max_results=max_results,
        file_pattern=file_pattern,
        exclude_dirs=exclude_dirs,
    )

    response = client.get_httpx_client().request(
        **kwargs,
    )

    return _build_response(client=client, response=response)


def sync(
    path: str,
    *,
    client: Union[Client],
    query: str,
    case_sensitive: Union[Unset, bool] = UNSET,
    context_lines: Union[Unset, int] = UNSET,
    max_results: Union[Unset, int] = UNSET,
    file_pattern: Union[Unset, str] = UNSET,
    exclude_dirs: Union[Unset, str] = UNSET,
) -> Union[ContentSearchResponse, ErrorResponse] | None:
    """Search for text content in files

     Searches for text content inside files using ripgrep. Returns matching lines with context.

    Args:
        path (str):
        query (str):
        case_sensitive (Union[Unset, bool]):
        context_lines (Union[Unset, int]):
        max_results (Union[Unset, int]):
        file_pattern (Union[Unset, str]):
        exclude_dirs (Union[Unset, str]):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[ContentSearchResponse, ErrorResponse]
    """

    return sync_detailed(
        path=path,
        client=client,
        query=query,
        case_sensitive=case_sensitive,
        context_lines=context_lines,
        max_results=max_results,
        file_pattern=file_pattern,
        exclude_dirs=exclude_dirs,
    ).parsed


async def asyncio_detailed(
    path: str,
    *,
    client: Union[Client],
    query: str,
    case_sensitive: Union[Unset, bool] = UNSET,
    context_lines: Union[Unset, int] = UNSET,
    max_results: Union[Unset, int] = UNSET,
    file_pattern: Union[Unset, str] = UNSET,
    exclude_dirs: Union[Unset, str] = UNSET,
) -> Response[Union[ContentSearchResponse, ErrorResponse]]:
    """Search for text content in files

     Searches for text content inside files using ripgrep. Returns matching lines with context.

    Args:
        path (str):
        query (str):
        case_sensitive (Union[Unset, bool]):
        context_lines (Union[Unset, int]):
        max_results (Union[Unset, int]):
        file_pattern (Union[Unset, str]):
        exclude_dirs (Union[Unset, str]):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[ContentSearchResponse, ErrorResponse]]
    """

    kwargs = _get_kwargs(
        path=path,
        query=query,
        case_sensitive=case_sensitive,
        context_lines=context_lines,
        max_results=max_results,
        file_pattern=file_pattern,
        exclude_dirs=exclude_dirs,
    )

    response = await client.get_async_httpx_client().request(**kwargs)

    return _build_response(client=client, response=response)


async def asyncio(
    path: str,
    *,
    client: Union[Client],
    query: str,
    case_sensitive: Union[Unset, bool] = UNSET,
    context_lines: Union[Unset, int] = UNSET,
    max_results: Union[Unset, int] = UNSET,
    file_pattern: Union[Unset, str] = UNSET,
    exclude_dirs: Union[Unset, str] = UNSET,
) -> Union[ContentSearchResponse, ErrorResponse] | None:
    """Search for text content in files

     Searches for text content inside files using ripgrep. Returns matching lines with context.

    Args:
        path (str):
        query (str):
        case_sensitive (Union[Unset, bool]):
        context_lines (Union[Unset, int]):
        max_results (Union[Unset, int]):
        file_pattern (Union[Unset, str]):
        exclude_dirs (Union[Unset, str]):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[ContentSearchResponse, ErrorResponse]
    """

    return (
        await asyncio_detailed(
            path=path,
            client=client,
            query=query,
            case_sensitive=case_sensitive,
            context_lines=context_lines,
            max_results=max_results,
            file_pattern=file_pattern,
            exclude_dirs=exclude_dirs,
        )
    ).parsed



import uuid
from typing import Dict, Union

from ..client.api.volumes.create_volume import asyncio as create_volume
from ..client.api.volumes.delete_volume import asyncio as delete_volume
from ..client.api.volumes.get_volume import asyncio as get_volume
from ..client.api.volumes.list_volumes import asyncio as list_volumes
from ..client.client import client
from ..client.models import Metadata, Volume, VolumeSpec
from ..client.types import UNSET


class VolumeCreateConfiguration:
    """Simplified configuration for creating volumes with default values."""

    def __init__(
        self,
        name: str | None = None,
        display_name: str | None = None,
        size: int | None = None,  # Size in MB
        region: str | None = None,  # AWS region
        template: str | None = None,  # Template
    ):
        self.name = name
        self.display_name = display_name
        self.size = size
        self.region = region
        self.template = template

    @classmethod
    def from_dict(cls, data: Dict[str, any]) -> "VolumeCreateConfiguration":
        return cls(
            name=data.get("name"),
            display_name=data.get("display_name"),
            size=data.get("size"),
            region=data.get("region"),
            template=data.get("template"),
        )


class VolumeInstance:
    def __init__(self, volume: Volume):
        self.volume = volume

    @property
    def metadata(self):
        return self.volume.metadata

    @property
    def spec(self):
        return self.volume.spec

    @property
    def status(self):
        return self.volume.status

    @property
    def name(self):
        return self.volume.metadata.name if self.volume.metadata else None

    @property
    def display_name(self):
        return self.volume.metadata.display_name if self.volume.metadata else None

    @property
    def size(self):
        return self.volume.spec.size if self.volume.spec else None

    @property
    def region(self):
        return self.volume.spec.region if self.volume.spec else None

    @property
    def template(self):
        return self.volume.spec.template if self.volume.spec else None

    @classmethod
    async def create(
        cls, config: Union[VolumeCreateConfiguration, Volume, Dict[str, any]]
    ) -> "VolumeInstance":
        # Generate default values
        default_name = f"volume-{uuid.uuid4().hex[:8]}"
        default_size = 1024  # 1GB in MB

        # Handle different configuration types
        if isinstance(config, Volume):
            volume = config
        elif isinstance(config, VolumeCreateConfiguration):
            volume = Volume(
                metadata=Metadata(
                    name=config.name or default_name,
                    display_name=config.display_name or config.name or default_name,
                ),
                spec=VolumeSpec(
                    size=config.size or default_size,
                    region=config.region or UNSET,
                    template=config.template or UNSET,
                ),
            )
        elif isinstance(config, dict):
            volume_config = VolumeCreateConfiguration.from_dict(config)
            volume = Volume(
                metadata=Metadata(
                    name=volume_config.name or default_name,
                    display_name=volume_config.display_name or volume_config.name or default_name,
                ),
                spec=VolumeSpec(
                    size=volume_config.size or default_size,
                    region=volume_config.region or UNSET,
                    template=volume_config.template or UNSET,
                ),
            )
        else:
            raise ValueError(
                f"Invalid config type: {type(config)}. Expected VolumeCreateConfiguration, Volume, or dict."
            )

        # Ensure required fields have defaults
        if not volume.metadata:
            volume.metadata = Metadata(name=default_name)
        if not volume.metadata.name:
            volume.metadata.name = default_name
        if not volume.spec:
            volume.spec = VolumeSpec(size=default_size)
        if not volume.spec.size:
            volume.spec.size = default_size

        response = await create_volume(client=client, body=volume)
        return cls(response)

    @classmethod
    async def get(cls, volume_name: str) -> "VolumeInstance":
        response = await get_volume(volume_name=volume_name, client=client)
        return cls(response)

    @classmethod
    async def list(cls) -> list["VolumeInstance"]:
        response = await list_volumes(client=client)
        return [cls(volume) for volume in response or []]

    @classmethod
    async def delete(cls, volume_name: str) -> Volume:
        response = await delete_volume(volume_name=volume_name, client=client)
        return response

    @classmethod
    async def create_if_not_exists(
        cls, config: Union[VolumeCreateConfiguration, Volume, Dict[str, any]]
    ) -> "VolumeInstance":
        """Create a volume if it doesn't exist, otherwise return existing."""
        try:
            return await cls.create(config)
        except Exception as e:
            # Check if it's a 409 conflict error (volume already exists)
            if (hasattr(e, "status_code") and e.status_code == 409) or (
                hasattr(e, "code") and e.code in [409, "VOLUME_ALREADY_EXISTS"]
            ):
                # Extract name from different configuration types
                if isinstance(config, VolumeCreateConfiguration):
                    name = config.name
                elif isinstance(config, dict):
                    name = config.get("name")
                elif isinstance(config, Volume):
                    name = config.metadata.name if config.metadata else None
                else:
                    name = None

                if not name:
                    raise ValueError("Volume name is required")

                volume_instance = await cls.get(name)
                return volume_instance
            raise e



from typing import Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="RevisionMetadata")


@_attrs_define
class RevisionMetadata:
    """Revision metadata

    Attributes:
        active (Union[Unset, bool]): Is the revision active
        canary (Union[Unset, bool]): Is the revision canary
        created_at (Union[Unset, str]): Revision created at
        created_by (Union[Unset, str]): Revision created by
        id (Union[Unset, str]): Revision ID
        previous_active (Union[Unset, bool]): Is the revision previous active
        status (Union[Unset, str]): Status of the revision
        traffic_percent (Union[Unset, int]): Percent of traffic to the revision
    """

    active: Union[Unset, bool] = UNSET
    canary: Union[Unset, bool] = UNSET
    created_at: Union[Unset, str] = UNSET
    created_by: Union[Unset, str] = UNSET
    id: Union[Unset, str] = UNSET
    previous_active: Union[Unset, bool] = UNSET
    status: Union[Unset, str] = UNSET
    traffic_percent: Union[Unset, int] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        active = self.active

        canary = self.canary

        created_at = self.created_at

        created_by = self.created_by

        id = self.id

        previous_active = self.previous_active

        status = self.status

        traffic_percent = self.traffic_percent

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if active is not UNSET:
            field_dict["active"] = active
        if canary is not UNSET:
            field_dict["canary"] = canary
        if created_at is not UNSET:
            field_dict["createdAt"] = created_at
        if created_by is not UNSET:
            field_dict["createdBy"] = created_by
        if id is not UNSET:
            field_dict["id"] = id
        if previous_active is not UNSET:
            field_dict["previousActive"] = previous_active
        if status is not UNSET:
            field_dict["status"] = status
        if traffic_percent is not UNSET:
            field_dict["trafficPercent"] = traffic_percent

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        if not src_dict:
            return None
        d = src_dict.copy()
        active = d.pop("active", UNSET)

        canary = d.pop("canary", UNSET)

        created_at = d.pop("createdAt", UNSET)

        created_by = d.pop("createdBy", UNSET)

        id = d.pop("id", UNSET)

        previous_active = d.pop("previousActive", UNSET)

        status = d.pop("status", UNSET)

        traffic_percent = d.pop("trafficPercent", UNSET)

        revision_metadata = cls(
            active=active,
            canary=canary,
            created_at=created_at,
            created_by=created_by,
            id=id,
            previous_active=previous_active,
            status=status,
            traffic_percent=traffic_percent,
        )

        revision_metadata.additional_properties = d
        return revision_metadata

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="ApplyEditRequest")


@_attrs_define
class ApplyEditRequest:
    """
    Attributes:
        code_edit (str):  Example: // Add world parameter
            function hello(world) {
              console.log('Hello', world);
            }.
        model (Union[Unset, str]):  Example: auto.
    """

    code_edit: str
    model: Union[Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        code_edit = self.code_edit

        model = self.model

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "codeEdit": code_edit,
            }
        )
        if model is not UNSET:
            field_dict["model"] = model

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        if not src_dict:
            return None
        d = src_dict.copy()
        code_edit = d.pop("codeEdit")

        model = d.pop("model", UNSET)

        apply_edit_request = cls(
            code_edit=code_edit,
            model=model,
        )

        apply_edit_request.additional_properties = d
        return apply_edit_request

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import asyncio
import sys

from blaxel.core.sandbox import SandboxCreateConfiguration, SandboxInstance


async def main():
    """Test sandbox snapshot feature."""
    try:
        # Test 1: Create sandbox with snapshot disabled using dict
        print("Test 1: Create sandbox with snapshot disabled (dict syntax)...")
        sandbox = await SandboxInstance.create({"snapshot_enabled": False})
        print(f"✅ Created sandbox with snapshot disabled: {sandbox.metadata.name}")
        print(f"   Snapshot enabled: {sandbox.spec.runtime.snapshot_enabled}")
        await asyncio.sleep(2)
        await sandbox.fs.ls("/")
        print("Waiting 20 seconds...")
        await asyncio.sleep(20)
        await sandbox.fs.ls("/")
        print("✅ Sandbox with snapshot disabled completed operations")

        # Test 2: Create sandbox with default snapshot settings (enabled by default)
        print("\nTest 2: Create sandbox with default snapshot settings...")
        sandbox2 = await SandboxInstance.create()
        print(f"✅ Created sandbox with default settings: {sandbox2.metadata.name}")
        print(f"   Snapshot enabled: {sandbox2.spec.runtime.snapshot_enabled}")
        await sandbox2.fs.ls("/")
        await sandbox2.fs.write("/test.txt", "Hello, world!")
        print("Waiting 30 seconds...")
        await asyncio.sleep(30)
        result = await sandbox2.fs.read("/test.txt")
        assert result == "Hello, world!", "File content mismatch"
        print("✅ Sandbox with default settings completed operations")
        print(f"   Directory contents: {result}")

        # Test 3: Create sandbox with snapshot disabled using SandboxCreateConfiguration
        print("\nTest 3: Create sandbox with snapshot disabled (config syntax)...")
        config = SandboxCreateConfiguration(
            name="sandbox-snapshot-disabled-config", snapshot_enabled=False
        )
        sandbox3 = await SandboxInstance.create(config)
        print(f"✅ Created sandbox with config: {sandbox3.metadata.name}")
        print(f"   Snapshot enabled: {sandbox3.spec.runtime.snapshot_enabled}")
        await sandbox3.fs.ls("/")
        print("✅ Config sandbox completed operations")

        # Cleanup
        print("\nCleaning up sandboxes...")
        await SandboxInstance.delete(sandbox.metadata.name)
        print(f"✅ Deleted sandbox: {sandbox.metadata.name}")
        await SandboxInstance.delete(sandbox2.metadata.name)
        print(f"✅ Deleted sandbox: {sandbox2.metadata.name}")
        await SandboxInstance.delete(sandbox3.metadata.name)
        print(f"✅ Deleted sandbox: {sandbox3.metadata.name}")

        print("\n🎉 All snapshot tests passed!")

    except Exception as e:
        print(f"❌ There was an error => {e}")
        import traceback

        traceback.print_exc()
        sys.exit(1)


if __name__ == "__main__":
    try:
        asyncio.run(main())
        print("✅ Tests completed successfully")
        sys.exit(0)
    except Exception as err:
        print(f"❌ There was an error => {err}")
        sys.exit(1)




from typing import TYPE_CHECKING, Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.port import Port


T = TypeVar("T", bound="SandboxDefinition")


@_attrs_define
class SandboxDefinition:
    """Sandbox definition for admin store operations

    Attributes:
        categories (Union[Unset, list[Any]]): Categories of the defintion
        coming_soon (Union[Unset, bool]): If the definition is coming soon
        description (Union[Unset, str]): Description of the defintion
        display_name (Union[Unset, str]): Display name of the definition
        enterprise (Union[Unset, bool]): If the definition is enterprise
        hidden (Union[Unset, bool]): If the definition is hidden
        icon (Union[Unset, str]): Icon of the definition
        image (Union[Unset, str]): Image of the Sandbox definition
        long_description (Union[Unset, str]): Long description of the defintion
        memory (Union[Unset, int]): Memory of the Sandbox definition in MB
        name (Union[Unset, str]): Name of the artifact
        ports (Union[Unset, list['Port']]): Set of ports for a resource
        tags (Union[Unset, str]): Tags of the definition
        url (Union[Unset, str]): URL of the definition
    """

    categories: Union[Unset, list[Any]] = UNSET
    coming_soon: Union[Unset, bool] = UNSET
    description: Union[Unset, str] = UNSET
    display_name: Union[Unset, str] = UNSET
    enterprise: Union[Unset, bool] = UNSET
    hidden: Union[Unset, bool] = UNSET
    icon: Union[Unset, str] = UNSET
    image: Union[Unset, str] = UNSET
    long_description: Union[Unset, str] = UNSET
    memory: Union[Unset, int] = UNSET
    name: Union[Unset, str] = UNSET
    ports: Union[Unset, list["Port"]] = UNSET
    tags: Union[Unset, str] = UNSET
    url: Union[Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        categories: Union[Unset, list[Any]] = UNSET
        if not isinstance(self.categories, Unset):
            categories = self.categories

        coming_soon = self.coming_soon

        description = self.description

        display_name = self.display_name

        enterprise = self.enterprise

        hidden = self.hidden

        icon = self.icon

        image = self.image

        long_description = self.long_description

        memory = self.memory

        name = self.name

        ports: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.ports, Unset):
            ports = []
            for componentsschemas_ports_item_data in self.ports:
                if type(componentsschemas_ports_item_data) is dict:
                    componentsschemas_ports_item = componentsschemas_ports_item_data
                else:
                    componentsschemas_ports_item = componentsschemas_ports_item_data.to_dict()
                ports.append(componentsschemas_ports_item)

        tags = self.tags

        url = self.url

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if categories is not UNSET:
            field_dict["categories"] = categories
        if coming_soon is not UNSET:
            field_dict["coming_soon"] = coming_soon
        if description is not UNSET:
            field_dict["description"] = description
        if display_name is not UNSET:
            field_dict["displayName"] = display_name
        if enterprise is not UNSET:
            field_dict["enterprise"] = enterprise
        if hidden is not UNSET:
            field_dict["hidden"] = hidden
        if icon is not UNSET:
            field_dict["icon"] = icon
        if image is not UNSET:
            field_dict["image"] = image
        if long_description is not UNSET:
            field_dict["longDescription"] = long_description
        if memory is not UNSET:
            field_dict["memory"] = memory
        if name is not UNSET:
            field_dict["name"] = name
        if ports is not UNSET:
            field_dict["ports"] = ports
        if tags is not UNSET:
            field_dict["tags"] = tags
        if url is not UNSET:
            field_dict["url"] = url

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        from ..models.port import Port

        if not src_dict:
            return None
        d = src_dict.copy()
        categories = cast(list[Any], d.pop("categories", UNSET))

        coming_soon = d.pop("coming_soon", UNSET)

        description = d.pop("description", UNSET)

        display_name = d.pop("displayName", UNSET)

        enterprise = d.pop("enterprise", UNSET)

        hidden = d.pop("hidden", UNSET)

        icon = d.pop("icon", UNSET)

        image = d.pop("image", UNSET)

        long_description = d.pop("longDescription", UNSET)

        memory = d.pop("memory", UNSET)

        name = d.pop("name", UNSET)

        ports = []
        _ports = d.pop("ports", UNSET)
        for componentsschemas_ports_item_data in _ports or []:
            componentsschemas_ports_item = Port.from_dict(componentsschemas_ports_item_data)

            ports.append(componentsschemas_ports_item)

        tags = d.pop("tags", UNSET)

        url = d.pop("url", UNSET)

        sandbox_definition = cls(
            categories=categories,
            coming_soon=coming_soon,
            description=description,
            display_name=display_name,
            enterprise=enterprise,
            hidden=hidden,
            icon=icon,
            image=image,
            long_description=long_description,
            memory=memory,
            name=name,
            ports=ports,
            tags=tags,
            url=url,
        )

        sandbox_definition.additional_properties = d
        return sandbox_definition

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import asyncio

from mcp.client.session import ClientSession
from mcp.client.streamable_http import streamablehttp_client

from blaxel import settings


async def run():
    """Run the completion client example."""
    url = f"{settings.run_url}/{settings.workspace}/functions/blaxel-search/mcp"

    async with streamablehttp_client(url, settings.headers) as (read, write, _):
        async with ClientSession(read, write) as session:
            # Initialize the connection
            await session.initialize()
            tools = await session.list_tools()
            print(f"Tools retrieved, number of tools: {len(tools.tools)}")

asyncio.run(run())


"""Tests for hash functions."""

from blaxel.core.common.internal import get_alphanumeric_limited_hash, get_global_unique_hash


def test_get_alphanumeric_limited_hash():
    """Test alphanumeric hash generation."""
    # Test basic functionality
    result = get_alphanumeric_limited_hash("charlou-function-blaxel-search", 48)
    assert isinstance(result, str)
    assert len(result) <= 48
    assert result.isalnum()

    # Test with specific known values
    result2 = get_alphanumeric_limited_hash("main-function-explorer-mcp", 48)
    assert isinstance(result2, str)
    assert len(result2) <= 48
    assert result2.isalnum()

    # Test consistency
    result3 = get_alphanumeric_limited_hash("charlou-function-blaxel-search", 48)
    assert result == result3, "Hash should be consistent"

    # Test different max sizes
    result_short = get_alphanumeric_limited_hash("test", 10)
    assert len(result_short) <= 10


def test_get_global_unique_hash():
    """Test global unique hash generation."""
    workspace = "test-workspace"
    type_name = "function"
    name = "test-function"

    result = get_global_unique_hash(workspace, type_name, name)
    assert isinstance(result, str)
    assert len(result) <= 48
    assert result.isalnum()

    # Test consistency
    result2 = get_global_unique_hash(workspace, type_name, name)
    assert result == result2, "Global unique hash should be consistent"

    # Test different inputs produce different hashes
    result3 = get_global_unique_hash("different-workspace", type_name, name)
    assert result != result3, "Different workspaces should produce different hashes"
    # Test different inputs produce different hashes
    result3 = get_global_unique_hash("different-workspace", type_name, name)
    assert result != result3, "Different workspaces should produce different hashes"



"""
This module provides the ApiKey class, which handles API key-based authentication for Blaxel.
"""

from typing import Generator

from httpx import Request, Response

from .types import BlaxelAuth


class ApiKey(BlaxelAuth):
    """
    ApiKey auth that authenticates requests using an API key.
    """

    def get_headers(self):
        """
        Retrieves the authentication headers containing the API key and workspace information.

        Returns:
            dict: A dictionary of headers with API key and workspace.
        """
        return {
            "X-Blaxel-Authorization": f"Bearer {self.credentials.api_key}",
            "X-Blaxel-Workspace": self.workspace_name,
        }

    def auth_flow(self, request: Request) -> Generator[Request, Response, None]:
        """
        Authenticates the request by adding API key and workspace headers.

        Parameters:
            request (Request): The HTTP request to authenticate.

        Yields:
            Request: The authenticated request.
        """
        request.headers["X-Blaxel-Authorization"] = f"Bearer {self.credentials.api_key}"
        request.headers["X-Blaxel-Workspace"] = self.workspace_name
        yield request

    @property
    def token(self):
        return self.credentials.api_key



from dataclasses import dataclass
from datetime import datetime
from typing import Any, Dict, List, Union

from ...client import errors
from ...client.api.compute.create_sandbox_preview import (
    asyncio as create_sandbox_preview,
)
from ...client.api.compute.create_sandbox_preview_token import (
    asyncio as create_sandbox_preview_token,
)
from ...client.api.compute.delete_sandbox_preview import (
    asyncio as delete_sandbox_preview,
)
from ...client.api.compute.delete_sandbox_preview_token import (
    asyncio as delete_sandbox_preview_token,
)
from ...client.api.compute.get_sandbox_preview import asyncio as get_sandbox_preview
from ...client.api.compute.list_sandbox_preview_tokens import (
    asyncio as list_sandbox_preview_tokens,
)
from ...client.api.compute.list_sandbox_previews import asyncio as list_sandbox_previews
from ...client.client import client
from ...client.models import (
    Preview,
    PreviewSpec,
    PreviewToken,
    PreviewTokenSpec,
    Sandbox,
)


@dataclass
class SandboxPreviewToken:
    """Represents a preview token with its value and expiration."""

    preview_token: PreviewToken

    @property
    def value(self) -> str:
        return self.preview_token.spec.token if self.preview_token.spec else ""

    @property
    def expires_at(self) -> datetime:
        return self.preview_token.spec.expires_at if self.preview_token.spec else datetime.now()


class SandboxPreviewTokens:
    """Manages preview tokens for a sandbox preview."""

    def __init__(self, preview: Preview):
        self.preview = preview

    @property
    def preview_name(self) -> str:
        return self.preview.metadata.name if self.preview.metadata else ""

    @property
    def resource_name(self) -> str:
        return self.preview.metadata.resource_name if self.preview.metadata else ""

    async def create(self, expires_at: datetime) -> SandboxPreviewToken:
        """Create a new preview token."""
        response: PreviewToken = await create_sandbox_preview_token(
            self.resource_name,
            self.preview_name,
            body=PreviewToken(
                spec=PreviewTokenSpec(
                    expires_at=to_utc_z(expires_at),
                )
            ),
            client=client,
        )
        return SandboxPreviewToken(response)

    async def list(self) -> List[SandboxPreviewToken]:
        """List all preview tokens."""
        response: List[PreviewToken] = await list_sandbox_preview_tokens(
            self.resource_name,
            self.preview_name,
            client=client,
        )
        return [SandboxPreviewToken(token) for token in response]

    async def delete(self, token_name: str) -> dict:
        """Delete a preview token."""
        response: PreviewToken = await delete_sandbox_preview_token(
            self.resource_name,
            self.preview_name,
            token_name,
            client=client,
        )
        return response


class SandboxPreview:
    """Represents a sandbox preview with its metadata and tokens."""

    def __init__(self, preview: Preview):
        self.preview = preview
        self.tokens = SandboxPreviewTokens(preview)

    @property
    def name(self) -> str:
        return self.preview.metadata.name if self.preview.metadata else ""

    @property
    def metadata(self) -> dict | None:
        return self.preview.metadata

    @property
    def spec(self) -> PreviewSpec | None:
        return self.preview.spec


class SandboxPreviews:
    """Manages sandbox previews."""

    def __init__(self, sandbox: Sandbox):
        self.sandbox = sandbox

    @property
    def sandbox_name(self) -> str:
        return self.sandbox.metadata.name if self.sandbox.metadata else ""

    async def list(self) -> List[SandboxPreview]:
        """List all previews for the sandbox."""
        response: List[Preview] = await list_sandbox_previews(
            self.sandbox_name,
            client=client,
        )
        return [SandboxPreview(preview) for preview in response]

    async def create(self, preview: Union[Preview, Dict[str, Any]]) -> SandboxPreview:
        """Create a new preview."""
        if isinstance(preview, dict):
            preview = Preview.from_dict(preview)

        response: Preview = await create_sandbox_preview(
            self.sandbox_name,
            body=preview,
            client=client,
        )
        return SandboxPreview(response)

    async def create_if_not_exists(self, preview: Union[Preview, Dict[str, Any]]) -> SandboxPreview:
        """Create a preview if it doesn't exist, otherwise return the existing one."""
        if isinstance(preview, dict):
            preview = Preview.from_dict(preview)

        preview_name = preview.metadata.name if preview.metadata else ""

        try:
            # Try to get existing preview
            existing_preview = await self.get(preview_name)
            return existing_preview
        except errors.UnexpectedStatus as e:
            # If 404, create the preview
            if e.status_code == 404:
                return await self.create(preview)
            # For any other error, re-raise
            raise e

    async def get(self, preview_name: str) -> SandboxPreview:
        """Get a specific preview by name."""
        response: Preview = await get_sandbox_preview(
            self.sandbox_name,
            preview_name,
            client=client,
        )
        return SandboxPreview(response)

    async def delete(self, preview_name: str) -> dict:
        """Delete a preview."""
        response: Preview = await delete_sandbox_preview(
            self.sandbox_name,
            preview_name,
            client=client,
        )
        return response


def to_utc_z(dt: datetime) -> str:
    """Convert datetime to UTC Z format string."""
    # Simple approach: format as ISO string and add Z
    iso_string = dt.isoformat()
    if iso_string.endswith("+00:00"):
        return iso_string.replace("+00:00", "Z")
    elif "T" in iso_string and not iso_string.endswith("Z"):
        return iso_string + "Z"
    return iso_string



from datetime import datetime
from typing import Any, Callable, Dict, List, TypeVar, Union

import httpx
from attrs import define as _attrs_define

from ..client.models import Port, Sandbox, SandboxLifecycle, VolumeAttachment
from ..client.types import UNSET
from .client.models.process_request import ProcessRequest
from .client.models.process_response import ProcessResponse


class SessionCreateOptions:
    def __init__(
        self,
        expires_at: datetime | None = None,
        response_headers: Dict[str, str] | None = None,
        request_headers: Dict[str, str] | None = None,
    ):
        self.expires_at = expires_at
        self.response_headers = response_headers or {}
        self.request_headers = request_headers or {}

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "SessionCreateOptions":
        expires_at = None
        if "expires_at" in data and data["expires_at"]:
            if isinstance(data["expires_at"], str):
                expires_at = datetime.fromisoformat(data["expires_at"].replace("Z", "+00:00"))
            elif isinstance(data["expires_at"], datetime):
                expires_at = data["expires_at"]

        return cls(
            expires_at=expires_at,
            response_headers=data.get("response_headers"),
            request_headers=data.get("request_headers"),
        )


class SessionWithToken:
    def __init__(self, name: str, url: str, token: str, expires_at: datetime):
        self.name = name
        self.url = url
        self.token = token
        self.expires_at = expires_at

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "SessionWithToken":
        expires_at = data["expires_at"]
        if isinstance(expires_at, str):
            expires_at = datetime.fromisoformat(expires_at.replace("Z", "+00:00"))

        return cls(
            name=data["name"],
            url=data["url"],
            token=data["token"],
            expires_at=expires_at,
        )


class VolumeBinding:
    """Volume binding configuration for sandbox."""

    def __init__(self, name: str, mount_path: str, read_only: bool | None = False):
        self.name = name
        self.mount_path = mount_path
        self.read_only = read_only or False

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "VolumeBinding":
        return cls(
            name=data["name"],
            mount_path=data["mount_path"],
            read_only=data.get("read_only", False),
        )


class SandboxConfiguration:
    def __init__(
        self,
        sandbox: Sandbox,
        force_url: str | None = None,
        headers: Dict[str, str] | None = None,
        params: Dict[str, str] | None = None,
    ):
        self.sandbox = sandbox
        self.force_url = force_url
        self.headers = headers or {}
        self.params = params or {}

    @property
    def metadata(self):
        return self.sandbox.metadata

    @property
    def status(self):
        return self.sandbox.status

    @property
    def spec(self):
        return self.sandbox.spec


class WatchEvent:
    def __init__(self, op: str, path: str, name: str, content: str | None = None):
        self.op = op
        self.path = path
        self.name = name
        self.content = content


class SandboxFilesystemFile:
    def __init__(self, path: str, content: str):
        self.path = path
        self.content = content

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "SandboxFilesystemFile":
        return cls(data["path"], data["content"])


class CopyResponse:
    def __init__(self, message: str, source: str, destination: str):
        self.message = message
        self.source = source
        self.destination = destination


class SandboxUpdateMetadata:
    """Configuration for updating sandbox metadata."""

    def __init__(
        self,
        labels: Dict[str, str] | None = None,
        display_name: str | None = None,
    ):
        self.labels = labels
        self.display_name = display_name


class SandboxCreateConfiguration:
    """Simplified configuration for creating sandboxes with default values."""

    def __init__(
        self,
        name: str | None = None,
        image: str | None = None,
        memory: int | None = None,
        ports: Union[List[Port], List[Dict[str, Any]]] | None = None,
        envs: List[Dict[str, str]] | None = None,
        volumes: Union[List[VolumeBinding], List[VolumeAttachment], List[Dict[str, Any]]] | None = None,
        ttl: str | None = None,
        expires: datetime | None = None,
        region: str | None = None,
        lifecycle: Union[SandboxLifecycle, Dict[str, Any]] | None = None,
        snapshot_enabled: bool | None = None,
    ):
        self.name = name
        self.image = image
        self.memory = memory
        self.ports = ports
        self.envs = envs
        self.volumes = volumes
        self.ttl = ttl
        self.expires = expires
        self.region = region
        self.lifecycle = lifecycle
        self.snapshot_enabled = snapshot_enabled

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "SandboxCreateConfiguration":
        expires = data.get("expires")
        if expires and isinstance(expires, str):
            expires = datetime.fromisoformat(expires.replace("Z", "+00:00"))

        lifecycle = data.get("lifecycle")
        if lifecycle and isinstance(lifecycle, dict):
            lifecycle = SandboxLifecycle.from_dict(lifecycle)

        return cls(
            name=data.get("name"),
            image=data.get("image"),
            memory=data.get("memory"),
            ports=data.get("ports"),
            envs=data.get("envs"),
            volumes=data.get("volumes"),
            ttl=data.get("ttl"),
            expires=expires,
            region=data.get("region"),
            lifecycle=lifecycle,
            snapshot_enabled=data.get("snapshot_enabled"),
        )

    def _normalize_ports(self) -> List[Port] | None:
        """Convert ports to Port objects with default protocol HTTP if not specified."""
        if not self.ports:
            return None

        port_objects = []
        for port in self.ports:
            if isinstance(port, Port):
                # If it's already a Port object, ensure protocol defaults to HTTP
                if port.protocol is UNSET or not port.protocol:
                    port.protocol = "HTTP"
                port_objects.append(port)
            elif isinstance(port, dict):
                # Convert dict to Port object with HTTP as default protocol
                port_dict = port.copy()
                if "protocol" not in port_dict or not port_dict["protocol"]:
                    port_dict["protocol"] = "HTTP"
                port_objects.append(Port.from_dict(port_dict))
            else:
                raise ValueError(f"Invalid port type: {type(port)}. Expected Port object or dict.")

        return port_objects

    def _normalize_envs(self) -> List[Dict[str, str]] | None:
        """Convert envs to list of dicts with name and value keys."""
        if not self.envs:
            return None

        env_objects = []
        for env in self.envs:
            if isinstance(env, dict):
                # Validate that the dict has the required keys
                if "name" not in env or "value" not in env:
                    raise ValueError(
                        f"Environment variable dict must have 'name' and 'value' keys: {env}"
                    )
                env_objects.append({"name": env["name"], "value": env["value"]})
            else:
                raise ValueError(
                    f"Invalid env type: {type(env)}. Expected dict with 'name' and 'value' keys."
                )

        return env_objects

    def _normalize_volumes(self) -> List[VolumeAttachment] | None:
        """Convert volumes to VolumeAttachment objects."""
        if not self.volumes:
            return None

        volume_objects = []
        for volume in self.volumes:
            if isinstance(volume, VolumeAttachment):
                volume_objects.append(volume)
            elif isinstance(volume, VolumeBinding):
                # Convert VolumeBinding to VolumeAttachment format
                volume_attachment = VolumeAttachment(
                    name=volume.name,
                    mount_path=volume.mount_path,
                    read_only=volume.read_only,
                )
                volume_objects.append(volume_attachment)
            elif isinstance(volume, dict):
                # Validate that the dict has the required keys
                if "name" not in volume or "mount_path" not in volume:
                    raise ValueError(
                        f"Volume binding dict must have 'name' and 'mount_path' keys: {volume}"
                    )
                if not isinstance(volume["name"], str) or not isinstance(volume["mount_path"], str):
                    raise ValueError(
                        f"Volume binding 'name' and 'mount_path' must be strings: {volume}"
                    )

                # Convert dict to VolumeAttachment object
                volume_attachment = VolumeAttachment(
                    name=volume["name"],
                    mount_path=volume["mount_path"],
                    read_only=volume.get("read_only", False),
                )
                volume_objects.append(volume_attachment)
            else:
                raise ValueError(
                    f"Invalid volume type: {type(volume)}. Expected VolumeAttachment, VolumeBinding, or dict with 'name' and 'mount_path' keys."
                )

        return volume_objects


@_attrs_define
class ProcessRequestWithLog(ProcessRequest):
    on_log: Callable[[str], None] | None = None


class ProcessResponseWithLog:
    """A process response with additional close functionality for stream management."""

    def __init__(self, process_response: ProcessResponse, close_func: Callable[[], None]):
        self._process_response = process_response
        self._close_func = close_func

    def close(self) -> None:
        """Close the log stream without terminating the process."""
        self._close_func()

    def __getattr__(self, name: str) -> Any:
        """Delegate attribute access to the underlying ProcessResponse."""
        return getattr(self._process_response, name)

    def __setattr__(self, name: str, value: Any) -> None:
        """Handle setting attributes, preserving special attributes."""
        if name.startswith("_") or name == "close":
            super().__setattr__(name, value)
        else:
            setattr(self._process_response, name, value)


class ResponseError(Exception):
    def __init__(self, response: httpx.Response):
        data_error = {}
        data = None
        if response.content:
            try:
                data = response.json()
                data_error = data
            except Exception:
                data = response.text
                data_error["response"] = data
        if response.status_code:
            data_error["status"] = response.status_code
        if response.reason_phrase:
            data_error["statusText"] = response.reason_phrase

        super().__init__(str(data_error))
        self.response = response
        self.data = data
        self.error = None


# -----------------------------
# Code Interpreter shared types
# -----------------------------

T = TypeVar("T")

# Generic output handler type
OutputHandler = Callable[[T], Any]


class OutputMessage:
    def __init__(self, text: str, timestamp: float | None, is_stderr: bool):
        self.text = text
        self.timestamp = timestamp
        self.is_stderr = is_stderr


class Result:
    def __init__(self, **kwargs: Any):
        for key, value in kwargs.items():
            setattr(self, key, value)


class ExecutionError:
    def __init__(self, name: str, value: Any, traceback: Any):
        self.name = name
        self.value = value
        self.traceback = traceback


class Logs:
    def __init__(self):
        self.stdout: List[str] = []
        self.stderr: List[str] = []


class Execution:
    def __init__(self):
        self.results: List[Result] = []
        self.logs: Logs = Logs()
        self.error: ExecutionError | None = None
        self.execution_count: int | None = None


class Context:
    def __init__(self, id: str):
        self.id = id

    @classmethod
    def from_json(cls, data: Dict[str, Any]) -> "Context":
        return cls(id=str(data.get("id") or data.get("context_id") or ""))



# ruff: noqa: E402

import asyncio

from dotenv import load_dotenv

load_dotenv()

from logging import getLogger

from pydantic_ai import Agent, CallToolsNode
from pydantic_ai.messages import ToolCallPart
from pydantic_ai.models import ModelSettings

from blaxel.pydantic import bl_model, bl_tools

logger = getLogger(__name__)

# MODEL = "gpt-4o-mini"
# MODEL = "claude-3-5-sonnet"
# MODEL = "xai-grok-beta"
# MODEL = "cohere-command-r-plus" # x -> tool call not supported
MODEL = "gemini-2-0-flash"
# MODEL = "deepseek-chat"
# MODEL = "mistral-large-latest"


async def main():
    tools = await bl_tools(["blaxel-search"])

    model = await bl_model(MODEL)

    input = "Search for the weather in San Francisco ?"
    agent = Agent(model=model, tools=tools, model_settings=ModelSettings(temperature=0))
    async with agent.iter(input) as agent_run:
        async for node in agent_run:
            if isinstance(node, CallToolsNode):
                for part in node.model_response.parts:
                    if isinstance(part, ToolCallPart):
                        logger.info(f"Tool call: {part}")
                    else:
                        logger.info(f"Response: {part}")


if __name__ == "__main__":
    asyncio.run(main())



from .autoload import autoload
from .env import env
from .internal import get_alphanumeric_limited_hash, get_global_unique_hash
from .settings import Settings, settings

__all__ = ["autoload", "Settings", "settings", "env", "get_alphanumeric_limited_hash", "get_global_unique_hash"]
